from __future__ import annotations

import hashlib
from collections.abc import Collection, Iterable
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

import tomlkit

from pdm.exceptions import PdmUsageError
from pdm.models.requirements import Requirement, parse_requirement
from pdm.utils import cd

if TYPE_CHECKING:
    from pdm.project.core import Project


class WorkspaceManager:
    """Manage workspace membership and workspace-derived project state."""

    def __init__(self, owner: Project) -> None:
        self.owner = owner

    @cached_property
    def project(self) -> Project | None:
        """Return the parent workspace project if the owner is a member."""
        if self.owner.is_global:
            return None
        for parent in self.owner.root.parents:
            if not parent.joinpath(self.owner.PYPROJECT_FILENAME).exists():
                continue
            project = self.owner.core.create_project(parent)
            if project.pyproject.settings.get("workspace") and project.workspace.has_member(self.owner.root):
                return project
        return None

    @property
    def parent_project(self) -> Project | None:
        """Return the nearest parent project that can host a workspace."""
        if self.owner.is_global:
            return None
        for parent in self.owner.root.parents:
            if parent.joinpath(self.owner.PYPROJECT_FILENAME).exists():
                return self.owner.core.create_project(parent)
        return None

    @property
    def is_root(self) -> bool:
        return bool(self.owner.pyproject.settings.get("workspace"))

    def iter_members(self) -> Iterable[Path]:
        """Iterate over resolved workspace member paths."""
        if self.owner.is_global:
            return
        workspace = self.owner.pyproject.settings.get("workspace") or {}
        members = workspace.get("members", [])
        seen: set[Path] = set()
        for pattern in members:
            paths = (
                self.owner.root.glob(pattern) if any(char in pattern for char in "*?[") else [self.owner.root / pattern]
            )
            for path in paths:
                path = path.resolve()
                if path in seen or not path.is_dir() or not path.joinpath(self.owner.PYPROJECT_FILENAME).is_file():
                    continue
                seen.add(path)
                yield path

    def has_member(self, path: str | Path) -> bool:
        candidate = Path(path).absolute().resolve()
        return any(candidate == member for member in self.iter_members())

    def iter_dependencies(self) -> Iterable[Requirement]:
        """Iterate over implicit editable requirements for workspace members."""
        workspace_project = self.project or self.owner
        if not workspace_project.workspace.is_root:
            return
        for member in workspace_project.workspace.iter_members():
            member_project = workspace_project.core.create_project(member)
            if not member_project.name or not member_project.is_distribution:
                continue
            with cd(workspace_project.root):
                req = parse_requirement("./" + member.relative_to(workspace_project.root).as_posix(), True)
                req.relocate(workspace_project.backend)  # type: ignore[attr-defined]
            req.name = member_project.name
            req.groups = ["default"]
            yield req

    def with_dependencies(
        self,
        requirements: Iterable[Requirement],
        *,
        exclude: Collection[str] | None = None,
    ) -> list[Requirement]:
        result = list(requirements)
        seen = {req.identify() for req in result}
        if exclude:
            seen.update(exclude)
        for req in self.iter_dependencies():
            if req.identify() in seen:
                continue
            result.append(req)
            seen.add(req.identify())
        return result

    def content_hash(self, algo: str = "sha256") -> str:
        """Return a project content hash including workspace members."""
        workspace_project = self.project or self.owner
        root_hash = workspace_project.pyproject.content_hash(algo)
        if not workspace_project.workspace.is_root:
            return root_hash

        hasher = hashlib.new(algo)
        hasher.update(root_hash.encode("utf-8"))
        root = workspace_project.root.resolve()
        for member in sorted(
            workspace_project.workspace.iter_members(),
            key=lambda path: path.relative_to(root).as_posix(),
        ):
            member_project = workspace_project.core.create_project(member)
            member_path = member.relative_to(root).as_posix()
            hasher.update(b"\0")
            hasher.update(member_path.encode("utf-8"))
            hasher.update(b"\0")
            hasher.update(member_project.pyproject.content_hash(algo).encode("utf-8"))
        return hasher.hexdigest()

    def add_member(self, path: str | Path, *, show_message: bool = True, dry_run: bool = False) -> None:
        if workspace_project := self.project:
            workspace_project.workspace.add_member(path, show_message=show_message, dry_run=dry_run)
            return
        member = Path(path).absolute().resolve()
        try:
            member_path = member.relative_to(self.owner.root.resolve()).as_posix()
        except ValueError as e:
            raise PdmUsageError(f"Workspace member {member} must be under {self.owner.root}") from e
        self.owner.pyproject.open_for_write()
        members = self.owner.pyproject.settings.setdefault("workspace", {}).setdefault("members", tomlkit.array())
        if member_path not in (str(item) for item in members):
            members.append(member_path)
            if not dry_run:
                self.owner.pyproject.write(show_message=show_message)

    def remove_member(self, path: str | Path, *, show_message: bool = True, dry_run: bool = False) -> bool:
        if workspace_project := self.project:
            return workspace_project.workspace.remove_member(path, show_message=show_message, dry_run=dry_run)
        member = Path(path).absolute().resolve()
        try:
            member_path = member.relative_to(self.owner.root.resolve()).as_posix()
        except ValueError:
            return False
        self.owner.pyproject.open_for_write()
        workspace = self.owner.pyproject.settings.get("workspace")
        if not workspace:
            return False
        members = workspace.get("members", [])
        removed = False
        for index in reversed([i for i, item in enumerate(members) if str(item) == member_path]):
            del members[index]
            removed = True
        if removed and not dry_run:
            self.owner.pyproject.write(show_message=show_message)
        return removed

    def maybe_add(self) -> Project | None:
        """Add the owner to the nearest parent workspace when applicable."""
        if self.project is not None:
            return self.project
        if project := self.parent_project:
            project.workspace.add_member(self.owner.root)
            self.__dict__.pop("project", None)
            self.owner._lockfile = None
            self.owner._environment = None
            return project
        return None



from __future__ import annotations

from types import SimpleNamespace

from pdm.models.working_set import EgglinkFinder, WorkingSet


def make_dist(name):
    return SimpleNamespace(metadata={"Name": name})


def test_egglink_finder_reads_link_target(tmp_path, mocker):
    site_packages = tmp_path / "site-packages"
    target = tmp_path / "target"
    site_packages.mkdir()
    target.mkdir()
    (site_packages / "demo.egg-link").write_text(f"{target}\n")
    distribution = mocker.sentinel.distribution
    finder = mocker.patch("pdm.models.working_set.im.MetadataPathFinder").return_value
    finder.find_distributions.return_value = [distribution]

    result = list(
        EgglinkFinder.find_distributions(
            SimpleNamespace(name="demo", path=[str(site_packages)]),
        )
    )

    assert result == [distribution]
    assert distribution.link_file == (site_packages / "demo.egg-link").absolute()


def test_egglink_finder_skips_missing_distribution(tmp_path, mocker):
    (tmp_path / "demo.egg-link").write_text(f"{tmp_path}\n")
    finder = mocker.patch("pdm.models.working_set.im.MetadataPathFinder").return_value
    finder.find_distributions.return_value = []

    result = list(EgglinkFinder.find_distributions(SimpleNamespace(name=None, path=[str(tmp_path)])))

    assert result == []


def test_working_set_prefers_owned_distribution_and_normalizes_names(mocker):
    owned = make_dist("Demo_Package")
    shared = make_dist("demo-package")
    unnamed = SimpleNamespace(metadata={})
    mocker.patch(
        "pdm.models.working_set.distributions",
        side_effect=[[owned, unnamed], [shared]],
    )

    working_set = WorkingSet(["owned", "owned"], ["shared", "shared"])

    assert list(working_set) == ["demo-package"]
    assert len(working_set) == 1
    assert working_set["demo-package"] is owned
    assert working_set.is_owned("demo-package")
    assert not working_set.is_owned("missing")


def test_working_set_uses_default_paths(mocker):
    distributions = mocker.patch("pdm.models.working_set.distributions", side_effect=[[], []])

    working_set = WorkingSet()

    assert len(working_set) == 0
    assert distributions.call_args_list[0].kwargs["path"]
    assert distributions.call_args_list[1].kwargs["path"] == []


def test_search_paths_ignores_missing_named_egglink(tmp_path):
    assert list(EgglinkFinder._search_paths("missing", [str(tmp_path)])) == []



from __future__ import annotations

import pytest

from pdm.exceptions import CandidateNotFound
from pdm.models.candidates import Candidate
from pdm.models.markers import EnvSpec
from pdm.models.repositories.lock import LockedRepository, Package
from pdm.models.requirements import parse_requirement
from pdm.models.specifiers import PySpecSet


def make_repository(mocker):
    repository = object.__new__(LockedRepository)
    repository.packages = {}
    repository.targets = []
    repository.env_spec = EnvSpec.current()
    repository.environment = mocker.Mock()
    repository.environment.project.backend.expand_line.side_effect = lambda value: value
    repository.environment.project.python_requires = PySpecSet(">=3.9")
    repository.environment.python_requires = PySpecSet(">=3.9")
    return repository


def make_candidate(requirement="demo", version="1.0"):
    return Candidate(parse_requirement(requirement), name="demo", version=version)


def test_get_dependencies_populates_candidate_metadata(mocker):
    repository = make_repository(mocker)
    locked = make_candidate(version="2.0")
    locked.requires_python = ">=3.9"
    key = ("demo", "2.0", None, False)
    repository.packages[key] = Package(locked, ["dep>=1"], "summary")
    candidate = make_candidate(version="2.0")
    candidate.name = None
    candidate.version = None
    candidate.requires_python = ""
    mocker.patch.object(repository, "_identify_candidate", return_value=key)

    metadata = repository._get_dependencies_from_lockfile(candidate)

    assert [requirement.as_line() for requirement in metadata.dependencies] == ["dep>=1"]
    assert metadata.requires_python == ">=3.9"
    assert metadata.summary == "summary"
    assert candidate.name == "demo"
    assert candidate.version == "2.0"


def test_get_dependencies_rejects_missing_dependency_data(mocker):
    repository = make_repository(mocker)
    candidate = make_candidate()
    key = repository._identify_candidate(candidate)
    repository.packages[key] = Package(candidate, None)

    with pytest.raises(CandidateNotFound, match="Missing dependencies"):
        repository._get_dependencies_from_lockfile(candidate)

    assert tuple(repository.dependency_generators()) == (repository._get_dependencies_from_lockfile,)


def test_matching_entries_for_named_requirement(mocker):
    repository = make_repository(mocker)
    demo = make_candidate()
    other = Candidate(parse_requirement("other"), name="other", version="1.0")
    repository.packages = {
        repository._identify_candidate(demo): Package(demo),
        repository._identify_candidate(other): Package(other),
    }

    assert list(repository._matching_entries(parse_requirement("demo"))) == [
        repository.packages[("demo", "1.0", None, False)]
    ]


def test_matching_entries_for_file_requirements(tmp_path, mocker):
    repository = make_repository(mocker)
    first_path = tmp_path / "first"
    second_path = tmp_path / "second"
    first_path.mkdir()
    second_path.mkdir()
    first = make_candidate(str(first_path))
    second = make_candidate(str(second_path))
    repository.packages = {
        ("demo", None, first_path.as_uri(), False): Package(first),
        ("demo", None, second_path.as_uri(), False): Package(second),
    }

    result = list(repository._matching_entries(parse_requirement(str(first_path))))

    assert result == [repository.packages[("demo", None, first_path.as_uri(), False)]]


def test_find_candidates_sets_file_requirement_name(mocker, tmp_path):
    repository = make_repository(mocker)
    path = tmp_path / "demo-1.0-py3-none-any.whl"
    locked = make_candidate(path.as_uri())
    requirement = parse_requirement(path.as_uri())
    mocker.patch.object(repository, "_matching_entries", return_value=[Package(locked)])
    mocker.patch.object(repository, "is_this_package", return_value=False)

    result = list(repository.find_candidates(requirement))

    assert result[0].name == "demo"
    assert requirement.name == "demo"


def test_find_candidates_prefers_current_project(mocker):
    repository = make_repository(mocker)
    candidate = make_candidate()
    mocker.patch.object(repository, "is_this_package", return_value=True)
    mocker.patch.object(repository, "make_this_candidate", return_value=candidate)
    matching = mocker.patch.object(repository, "_matching_entries")

    assert list(repository.find_candidates(parse_requirement("demo"))) == [candidate]
    matching.assert_not_called()


def test_evaluate_candidates_filters_packages(mocker):
    repository = make_repository(mocker)
    repository.environment.project.split_extras_groups.return_value = (["extra"], ["dev"])
    repository.environment.project.pyproject.resolution = {"excludes": ["excluded"]}

    def package(name, *, marker_matches=True, group_matches=True, groups=None):
        candidate = make_candidate(name)
        candidate.req.groups = groups or []
        candidate.req.marker = mocker.Mock()
        candidate.req.marker.matches.return_value = marker_matches
        group_marker = mocker.Mock()
        group_marker.evaluate.return_value = group_matches
        return Package(candidate, marker=group_marker)

    included = package("included", groups=["default"])
    repository.packages = {
        ("excluded", "1.0", None, False): package("excluded"),
        ("env-mismatch", "1.0", None, False): package("env-mismatch", marker_matches=False),
        ("group-marker", "1.0", None, False): package("group-marker", group_matches=False),
        ("wrong-group", "1.0", None, False): package("wrong-group", groups=["dev"]),
        ("included", "1.0", None, False): included,
    }

    result = list(repository.evaluate_candidates(["default"]))

    assert result == [included]


def test_evaluate_candidates_can_skip_environment_markers(mocker):
    repository = make_repository(mocker)
    repository.environment.project.split_extras_groups.return_value = ([], [])
    repository.environment.project.pyproject.resolution = {}
    candidate = make_candidate()
    candidate.req.marker = mocker.Mock()
    candidate.req.marker.matches.return_value = False
    package = Package(candidate)
    repository.packages = {("demo", "1.0", None, False): package}

    assert list(repository.evaluate_candidates(["default"], evaluate_markers=False)) == [package]


def test_merge_result_adds_and_merges_packages(mocker):
    repository = make_repository(mocker)
    env_spec = EnvSpec.current()
    existing = make_candidate("demo; python_version < '3.12'")
    existing.req.groups = ["default"]
    existing.hashes = [{"hash": "sha256:old"}]
    incoming = make_candidate("demo; python_version >= '3.10'")
    incoming.req.groups = ["dev"]
    incoming.hashes = [{"hash": "sha256:old"}, {"hash": "sha256:new"}]
    new_candidate = Candidate(parse_requirement("new"), name="new", version="1.0")
    repository.add_package(Package(existing))
    repository.__dict__["all_candidates"] = {"stale": []}

    repository.merge_result(env_spec, [Package(incoming), Package(new_candidate)])

    merged = repository.packages[repository._identify_candidate(existing)].candidate
    assert set(merged.req.groups) == {"default", "dev"}
    assert merged.hashes == [{"hash": "sha256:old"}, {"hash": "sha256:new"}]
    assert repository._identify_candidate(new_candidate) in repository.packages
    assert env_spec in repository.targets
    assert "all_candidates" not in repository.__dict__


def test_get_hashes_returns_candidate_hashes(mocker):
    repository = make_repository(mocker)
    candidate = make_candidate()
    candidate.hashes = [{"hash": "sha256:value"}]

    assert repository.get_hashes(candidate) is candidate.hashes



from __future__ import annotations

import pytest

from pdm._types import RepositoryConfig
from pdm.exceptions import PdmException
from pdm.models.auth import Keyring, PdmBasicAuth
from pdm.termui import Verbosity


def test_auth_selects_closest_source(mocker):
    sources = [
        RepositoryConfig("pypi", "root", url="https://example.org/simple"),
        RepositoryConfig(
            "pypi",
            "private",
            url="https://example.org/private/simple",
            username="user",
            password="pass",
        ),
    ]
    auth = PdmBasicAuth(mocker.Mock(), sources)

    credentials, index = auth._get_auth_from_index_url("https://example.org/private/packages/demo")

    assert credentials == ("user", "pass")
    assert index == "https://example.org/private/simple/"
    assert auth._selected_source is sources[1]


def test_auth_handles_empty_and_unmatched_sources(mocker):
    auth = PdmBasicAuth(mocker.Mock(), [])
    assert auth._get_auth_from_index_url("https://example.org/simple") == (None, None)

    auth.sources = [RepositoryConfig("pypi", "other", url="https://other.org/simple")]
    assert auth._get_auth_from_index_url("https://example.org/simple") == (None, None)


def test_auth_uses_source_keyring_credentials(mocker):
    source = RepositoryConfig("pypi", "private", url="https://example.org/simple")
    source.populate_keyring_auth = mocker.Mock(
        side_effect=lambda: (
            setattr(source, "username", "keyring-user"),
            setattr(source, "password", "keyring-pass"),
        )
    )
    auth = PdmBasicAuth(mocker.Mock(), [source])
    auth._selected_source = source
    mocker.patch(
        "unearth.auth.MultiDomainBasicAuth._get_new_credentials",
        return_value=(None, None),
    )

    result = auth._get_new_credentials("https://example.org", allow_keyring=True)

    assert result == ("keyring-user", "keyring-pass")
    assert auth._selected_source is None


def test_auth_requires_verbose_mode_for_password_prompt(mocker):
    ui = mocker.Mock(verbosity=Verbosity.NORMAL)
    auth = PdmBasicAuth(ui, [])

    with pytest.raises(PdmException, match="rerun the command with `-v`"):
        auth._prompt_for_password("example.org")


def test_auth_verbose_password_prompt_delegates(mocker):
    ui = mocker.Mock(verbosity=Verbosity.DETAIL)
    auth = PdmBasicAuth(ui, [])
    parent = mocker.patch(
        "unearth.auth.MultiDomainBasicAuth._prompt_for_password",
        return_value=("user", "pass", True),
    )

    assert auth._prompt_for_password("example.org", "user") == ("user", "pass", True)
    parent.assert_called_once_with("example.org", "user")


def test_auth_warns_when_keyring_is_unavailable(mocker):
    ui = mocker.Mock()
    auth = PdmBasicAuth(ui, [])
    mocker.patch("pdm.models.auth.get_keyring_provider", return_value=None)
    parent = mocker.patch(
        "unearth.auth.MultiDomainBasicAuth._should_save_password_to_keyring",
        return_value=False,
    )

    assert auth._should_save_password_to_keyring() is False
    ui.info.assert_called_once()
    parent.assert_called_once_with()


def test_auth_does_not_warn_when_keyring_is_available(mocker):
    ui = mocker.Mock()
    auth = PdmBasicAuth(ui, [])
    mocker.patch("pdm.models.auth.get_keyring_provider", return_value=mocker.sentinel.provider)
    mocker.patch(
        "unearth.auth.MultiDomainBasicAuth._should_save_password_to_keyring",
        return_value=True,
    )

    assert auth._should_save_password_to_keyring() is True
    ui.info.assert_not_called()


@pytest.mark.parametrize("operation", ["get", "save", "delete"])
def test_keyring_disables_itself_after_provider_error(mocker, operation):
    provider = mocker.Mock()
    keyring = Keyring()
    keyring.provider = provider
    keyring.enabled = True
    method = getattr(provider, f"{operation}_auth_info")
    method.side_effect = RuntimeError("keyring failed")

    if operation == "get":
        result = keyring.get_auth_info("service", "user")
    else:
        result = getattr(keyring, f"{operation}_auth_info")(
            "service", "user", *(("pass",) if operation == "save" else ())
        )

    assert result in (None, False)
    assert keyring.enabled is False


def test_disabled_keyring_does_not_delete_credentials(mocker):
    keyring = Keyring()
    keyring.provider = mocker.Mock()
    keyring.enabled = False

    assert keyring.delete_auth_info("service", "user") is False
    keyring.provider.delete_auth_info.assert_not_called()



import argparse
import re
import sys
from typing import cast

from pdm import termui
from pdm.cli.commands.base import BaseCommand
from pdm.cli.filters import GroupSelection
from pdm.cli.hooks import HookManager
from pdm.cli.options import (
    config_setting_option,
    groups_group,
    lock_strategy_group,
    lockfile_option,
    no_isolation_option,
    override_option,
    skip_option,
)
from pdm.models.markers import EnvSpec
from pdm.models.specifiers import PySpecSet
from pdm.project import Project
from pdm.utils import convert_to_datetime


class Command(BaseCommand):
    """Resolve and lock dependencies"""

    arguments = (
        *BaseCommand.arguments,
        lockfile_option,
        no_isolation_option,
        config_setting_option,
        override_option,
        skip_option,
        groups_group,
        lock_strategy_group,
    )

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--refresh",
            action="store_true",
            help="Refresh the lock input metadata and file hashes in the lock file",
        )

        parser.add_argument(
            "--check",
            action="store_true",
            help="Check if the lock file is up to date and quit",
        )
        parser.add_argument(
            "--update-reuse",
            action="store_const",
            dest="update_strategy",
            default="all",
            const="reuse",
            help="Reuse pinned versions already present in lock file if possible",
        )
        parser.add_argument(
            "--update-reuse-installed",
            action="store_const",
            dest="update_strategy",
            const="reuse-installed",
            help="Reuse installed packages if possible",
        )
        parser.add_argument(
            "--exclude-newer",
            help="Exclude packages newer than the given UTC date in format `YYYY-MM-DD[THH:MM:SSZ]` "
            "or relative time `N{d|h|w}`",
            type=convert_to_datetime,
        )

        target_group = parser.add_argument_group("Lock Target")
        target_group.add_argument("--python", help="The Python range to lock for. E.g. `>=3.9`, `==3.12.*`")
        target_group.add_argument(
            "--platform",
            help="The platform to lock for. E.g. `windows`, `linux`, `macos`, `manylinux_2_17_x86_64`. "
            "See docs for available choices: http://pdm-project.org/en/latest/usage/lock-targets/",
        )
        target_group.add_argument(
            "--implementation",
            help="The Python implementation to lock for. E.g. `cpython`, `pypy`, `pyston`",
        )
        target_group.add_argument("--append", action="store_true", help="Append the result to the current lock file")

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        from pdm.cli import actions

        if options.check:
            strategy = actions.check_lockfile(project, False)
            if strategy:
                project.core.ui.echo(
                    f"[error]{termui.Emoji.FAIL}[/] Lockfile is [error]out of date[/].",
                    err=True,
                    verbosity=termui.Verbosity.DETAIL,
                )
                sys.exit(1)
            else:
                project.core.ui.echo(
                    f"[success]{termui.Emoji.SUCC}[/] Lockfile is [success]up to date[/].",
                    err=True,
                    verbosity=termui.Verbosity.DETAIL,
                )
                sys.exit(0)
        selection = GroupSelection.from_options(project, options)
        strategy = options.update_strategy
        if options.exclude_newer:
            # cli args override pyproject config if present
            project.core.state.exclude_newer = options.exclude_newer
        if project.core.state.exclude_newer:
            strategy = "all"
            if strategy != options.update_strategy:
                project.core.ui.info("--exclude-newer is set, forcing --update-all")
        env_spec: EnvSpec | None = None
        if any([options.python, options.platform, options.implementation]):
            replace_dict = {}
            if options.python:
                if re.match(r"[\d.]+", options.python):
                    options.python = f">={options.python}"
                replace_dict["requires_python"] = PySpecSet(options.python)
            if options.platform:
                replace_dict["platform"] = options.platform
            if options.implementation:
                replace_dict["implementation"] = options.implementation
            env_spec = project.environment.allow_all_spec.replace(**replace_dict)

        actions.do_lock(
            project,
            refresh=options.refresh,
            strategy=cast(str, strategy),
            groups=selection.all(),
            strategy_change=options.strategy_change,
            hooks=HookManager(project, options.skip),
            env_spec=env_spec,
            append=options.append,
        )



from types import SimpleNamespace
from unittest.mock import ANY

import pytest

from pdm import termui
from pdm.cli.commands import self_cmd


def mock_distribution(metadata, entry_points=()):
    entry_points = (SimpleNamespace(group=ep) for ep in entry_points)
    return SimpleNamespace(metadata=metadata, entry_points=entry_points)


DISTRIBUTIONS = {
    "foo": mock_distribution({"Name": "foo", "Version": "1.0.0", "Summary": "Foo package"}, ["pdm.plugin"]),
    "bar": mock_distribution({"Name": "bar", "Version": "2.0.0", "Summary": "Bar package"}, ["pdm"]),
    "baz": mock_distribution({"Name": "baz", "Version": "3.0.0", "Summary": "Baz package"}),
}


@pytest.fixture()
def mock_pip(mocker):
    mocked = mocker.patch("pdm.cli.commands.self_cmd.run_pip")
    return mocked


@pytest.fixture()
def mock_all_distributions(mocker):
    mocker.patch("pdm.models.working_set.WorkingSet", return_value=DISTRIBUTIONS)


@pytest.fixture()
def mock_latest_pdm_version(mocker):
    return mocker.patch(
        "pdm.cli.actions.get_latest_pdm_version_from_pypi",
    )


@pytest.mark.usefixtures("mock_all_distributions")
def test_self_list(pdm):
    result = pdm(["self", "list"])
    assert result.exit_code == 0, result.stderr
    packages = [line.split()[0] for line in result.stdout.splitlines()]
    assert packages == ["bar", "baz", "foo"]


@pytest.mark.usefixtures("mock_all_distributions")
def test_self_list_plugins(pdm):
    result = pdm(["self", "list", "--plugins"])
    assert result.exit_code == 0, result.stderr
    packages = [line.split()[0] for line in result.stdout.splitlines()]
    assert packages == ["bar", "foo"]


def test_self_add(pdm, mock_pip):
    result = pdm(["self", "add", "foo"])
    assert result.exit_code == 0, result.stderr
    mock_pip.assert_called_with(ANY, ["install", "foo"])

    result = pdm(["self", "add", "--pip-args", "--force-reinstall --upgrade", "foo"])
    assert result.exit_code == 0, result.stderr
    mock_pip.assert_called_with(ANY, ["install", "--force-reinstall", "--upgrade", "foo"])


def test_self_remove(pdm, mock_pip, mocker, monkeypatch):
    from rich import get_console

    console = get_console()

    def _mock_resolve(packages):
        return ["demo", "pytz"] if "demo" in packages else packages

    mocker.patch.object(
        self_cmd.RemoveCommand,
        "_resolve_dependencies_to_remove",
        side_effect=_mock_resolve,
    )
    mocker.patch.object(console, "is_interactive", True)

    result = pdm(["self", "remove", "foo"])
    assert result.exit_code != 0

    result = pdm(["self", "remove", "-y", "demo"])
    assert result.exit_code == 0, result.stderr
    mock_pip.assert_called_with(ANY, ["uninstall", "-y", "demo", "pytz"])

    with monkeypatch.context() as m:
        m.setenv("PDM_NON_INTERACTIVE", "1")
        result = pdm(["self", "remove", "demo"])
        assert result.exit_code == 0, result.stderr
        mock_pip.assert_called_with(ANY, ["uninstall", "-y", "demo", "pytz"])

    result = pdm(["-n", "self", "remove", "demo"])
    assert result.exit_code == 0, result.stderr
    mock_pip.assert_called_with(ANY, ["uninstall", "-y", "demo", "pytz"])


@pytest.mark.parametrize(
    "args,expected",
    [
        (["self", "update"], ["install", "--upgrade", "--upgrade-strategy", "eager", "pdm[locked]==99.0.0"]),
        (["self", "update", "--pre"], ["install", "--upgrade", "--upgrade-strategy", "eager", "pdm[locked]==99.0.1b1"]),
        (
            ["self", "update", "--head"],
            ["install", "--upgrade", "--upgrade-strategy", "eager", f"pdm[locked] @ git+{self_cmd.PDM_REPO}@main"],
        ),
    ],
)
def test_self_update(pdm, mock_pip, mock_latest_pdm_version, args, expected):
    def mocked_latest_version(project, pre):
        return "99.0.1b1" if pre else "99.0.0"

    mock_latest_pdm_version.side_effect = mocked_latest_version

    result = pdm(args)
    assert result.exit_code == 0, result.stderr
    mock_pip.assert_called_with(ANY, expected)


def test_self_update_already_latest(pdm, mock_pip, mock_latest_pdm_version):
    mock_latest_pdm_version.return_value = "0.0.0"

    result = pdm(["self", "update"])
    assert result.exit_code == 0, result.stderr
    assert "Already up-to-date" in result.stdout
    mock_pip.assert_not_called()


@pytest.mark.parametrize("use_uv", [False, True])
def test_run_pip_builds_command(project, mocker, use_uv):
    project.project_config["use_uv"] = use_uv
    project.core.uv_cmd = ["uv"]
    completed = mocker.patch("pdm.cli.commands.self_cmd.subprocess.run").return_value
    if not use_uv:
        environment = mocker.Mock(pip_command=["python", "-m", "pip"])
        bare_environment = mocker.patch("pdm.environments.BareEnvironment", return_value=environment)

    result = self_cmd.run_pip(project, ["install", "--upgrade-strategy", "eager", "demo"])

    assert result is completed
    if use_uv:
        expected = ["uv", "pip", "install", "demo", "--python", self_cmd.sys.executable]
    else:
        expected = ["python", "-m", "pip", "install", "--upgrade-strategy", "eager", "demo"]
        bare_environment.assert_called_once_with(project)
        assert project.environment is environment
    self_cmd.subprocess.run.assert_called_once_with(
        expected,
        stdout=self_cmd.subprocess.PIPE,
        stderr=self_cmd.subprocess.STDOUT,
        check=True,
        text=True,
    )


def test_self_list_without_plugins_exits(pdm, mocker):
    mocker.patch("pdm.cli.commands.self_cmd.list_distributions", return_value=[])

    result = pdm(["self", "list", "--plugins"])

    assert result.exit_code == 1
    assert "No plugin is installed" in result.stderr


@pytest.mark.parametrize("command", ["add", "remove", "update"])
def test_self_commands_report_pip_failure(pdm, mocker, command):
    error = self_cmd.subprocess.CalledProcessError(1, ["pip"], output="pip failed")
    mocker.patch("pdm.cli.commands.self_cmd.run_pip", side_effect=error)
    if command == "remove":
        mocker.patch.object(self_cmd.RemoveCommand, "_resolve_dependencies_to_remove", return_value=["demo"])
        args = ["self", command, "-y", "demo"]
    elif command == "update":
        mocker.patch("pdm.cli.actions.get_latest_pdm_version_from_pypi", return_value="99.0.0")
        args = ["self", command]
    else:
        args = ["self", command, "demo"]

    result = pdm(args)

    assert result.exit_code == 1
    assert "pip failed" in result.stderr


def test_self_remove_resolves_orphan_dependencies(mocker):
    root = mock_distribution({"Name": "root", "Version": "1.0"}, ())
    root.requires = ["dep", "pdm"]
    root.version = "1.0"
    dep = mock_distribution({"Name": "dep", "Version": "1.0"}, ())
    dep.requires = []
    dep.version = "1.0"
    pdm_dist = mock_distribution({"Name": "pdm", "Version": "1.0"}, ())
    pdm_dist.requires = []
    pdm_dist.version = "1.0"
    working_set = {"root": root, "dep": dep, "pdm": pdm_dist}
    mocker.patch("pdm.models.working_set.WorkingSet", return_value=working_set)

    result = self_cmd.RemoveCommand()._resolve_dependencies_to_remove(["root", "missing"])

    assert result == ["dep", "root"]


def test_self_remove_can_be_cancelled(pdm, mocker):
    mocker.patch.object(self_cmd.RemoveCommand, "_resolve_dependencies_to_remove", return_value=["demo"])
    mocker.patch.object(termui, "confirm", return_value=False)
    run_pip = mocker.patch("pdm.cli.commands.self_cmd.run_pip")

    result = pdm(["self", "remove", "demo"])

    assert result.exit_code == 0
    run_pip.assert_not_called()


def test_self_remove_without_matching_packages(pdm, mocker):
    mocker.patch.object(self_cmd.RemoveCommand, "_resolve_dependencies_to_remove", return_value=[])

    result = pdm(["self", "remove", "missing"])

    assert result.exit_code == 1
    assert "No package to remove" in result.stderr


def test_self_command_without_subcommand_shows_help(pdm):
    result = pdm(["self"])

    assert result.exit_code == 0
    assert "Manage the PDM program itself" in result.stdout


def test_self_zipapp_only_registers_list_command(mocker):
    parser = self_cmd.argparse.ArgumentParser()
    mocker.patch("pdm.cli.commands.self_cmd.is_in_zipapp", return_value=True)
    list_register = mocker.patch.object(self_cmd.ListCommand, "register_to")
    add_register = mocker.patch.object(self_cmd.AddCommand, "register_to")

    self_cmd.Command().add_arguments(parser)

    list_register.assert_called_once()
    add_register.assert_not_called()


def test_run_pip_uv_without_upgrade_strategy(project, mocker):
    project.project_config["use_uv"] = True
    project.core.uv_cmd = ["uv"]
    run = mocker.patch("pdm.cli.commands.self_cmd.subprocess.run")

    self_cmd.run_pip(project, ["install", "demo"])

    assert run.call_args.args[0] == ["uv", "pip", "install", "demo", "--python", self_cmd.sys.executable]



from datetime import datetime, timezone
from pathlib import Path
from unittest.mock import ANY

import pytest
from unearth import Link

from pdm.cli import actions
from pdm.exceptions import PdmUsageError
from pdm.models.requirements import parse_requirement
from pdm.models.specifiers import PySpecSet
from pdm.project.lockfile import FLAG_CROSS_PLATFORM
from pdm.project.lockfile.base import Compatibility, LockInputsState
from pdm.project.lockfile.freshness import lock_inputs_match
from pdm.signals import pre_lock
from pdm.utils import parse_version
from tests import FIXTURES


def test_lock_command(project, pdm, mocker):
    m = mocker.patch.object(actions, "do_lock")
    pdm(["lock"], obj=project)
    m.assert_called_with(
        project,
        refresh=False,
        groups=["default"],
        hooks=ANY,
        strategy_change=None,
        strategy="all",
        append=False,
        env_spec=None,
    )


class FakeResolver:
    def __init__(self, *args, **kwargs):
        pass

    def resolve(self):
        return [], set()


def make_workspace_member(project, core):
    project.pyproject.settings["workspace"] = {"members": ["packages/*"]}
    project.pyproject.write()
    member_path = project.root / "packages" / "foo"
    member_path.mkdir(parents=True)
    member_path.joinpath("pyproject.toml").write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = ["requests"]\n',
        encoding="utf-8",
    )
    return core.create_project(member_path, global_config=project.global_config.config_file.as_posix())


def make_workspace_members(project, core, members):
    project.pyproject.settings["workspace"] = {"members": ["packages/*"]}
    project.pyproject.write()
    for name, dependencies in members.items():
        member_path = project.root / "packages" / name
        member_path.mkdir(parents=True)
        dependency_lines = ", ".join(f'"{dependency}"' for dependency in dependencies)
        member_path.joinpath("pyproject.toml").write_text(
            f'[project]\nname = "{name}"\nversion = "0.1.0"\ndependencies = [{dependency_lines}]\n',
            encoding="utf-8",
        )
    return core.create_project(project.root, global_config=project.global_config.config_file.as_posix())


def capture_pre_lock_requirements():
    captured = {}

    def capture_requirements(sender, requirements, **kwargs):
        captured["requirements"] = requirements

    pre_lock.connect(capture_requirements, weak=False)
    return captured, capture_requirements


def enable_lock_inputs(project):
    project.pyproject.open_for_write()
    project.pyproject.settings.setdefault("resolution", {})["lock_inputs"] = True
    project.pyproject.write(show_message=False)


def test_do_lock_adds_workspace_members_to_explicit_member_requirements(project, core, mocker):
    member_project = make_workspace_member(project, core)
    mocker.patch.object(member_project, "get_resolver", return_value=FakeResolver)
    captured, receiver = capture_pre_lock_requirements()

    try:
        actions.do_lock(
            member_project,
            requirements=[parse_requirement("requests")],
            groups=["default"],
            dry_run=True,
        )
    finally:
        pre_lock.disconnect(receiver)

    requirements = captured["requirements"]
    assert [req.identify() for req in requirements] == ["requests", "foo"]
    assert requirements[-1].editable
    assert requirements[-1].str_path == "./packages/foo"
    assert member_project.lockfile._path == project.root / "pdm.lock"


def test_do_lock_adds_workspace_members_to_member_resolved_requirements(project, core, mocker):
    member_project = make_workspace_member(project, core)
    mocker.patch.object(member_project, "get_resolver", return_value=FakeResolver)
    captured, receiver = capture_pre_lock_requirements()

    try:
        actions.do_lock(member_project, groups=["default"], dry_run=True)
    finally:
        pre_lock.disconnect(receiver)

    requirements = captured["requirements"]
    assert [req.identify() for req in requirements] == ["requests", "foo"]
    assert requirements[-1].editable
    assert requirements[-1].str_path == "./packages/foo"
    assert member_project.lockfile._path == project.root / "pdm.lock"


def test_lock_workspace_members_depend_on_each_other(project, core, repository):
    workspace_project = make_workspace_members(project, core, {"foo": ["bar"], "bar": []})

    actions.do_lock(workspace_project)

    locked_candidates = workspace_project.get_locked_repository().candidates
    assert locked_candidates["foo"].req.editable
    assert locked_candidates["bar"].req.editable
    assert locked_candidates["foo"].req.str_path == "./packages/foo"
    assert locked_candidates["bar"].req.str_path == "./packages/bar"


@pytest.mark.usefixtures("repository")
@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_hash_tracks_workspace_member_pyproject(project, core, lock_format):
    project.project_config["lock.format"] = lock_format
    workspace_project = make_workspace_members(project, core, {"foo": []})

    actions.do_lock(workspace_project)

    assert workspace_project.is_lockfile_hash_match()
    assert workspace_project.is_lockfile_fresh()

    member_pyproject = workspace_project.root / "packages" / "foo" / "pyproject.toml"
    member_pyproject.write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = ["requests"]\n',
        encoding="utf-8",
    )
    fresh_project = core.create_project(
        workspace_project.root, global_config=workspace_project.global_config.config_file.as_posix()
    )
    fresh_member_project = core.create_project(
        member_pyproject.parent, global_config=workspace_project.global_config.config_file.as_posix()
    )

    assert not fresh_project.is_lockfile_hash_match()
    assert not fresh_member_project.is_lockfile_hash_match()
    assert not fresh_project.is_lockfile_fresh()
    assert not fresh_member_project.is_lockfile_fresh()

    actions.do_lock(fresh_project)

    assert fresh_project.is_lockfile_hash_match()


@pytest.mark.usefixtures("repository")
@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_inputs_track_workspace_member_version(project, core, pdm, lock_format):
    project.project_config["lock.format"] = lock_format
    workspace_project = make_workspace_members(project, core, {"foo": []})
    enable_lock_inputs(workspace_project)
    actions.do_lock(workspace_project)

    member_pyproject = workspace_project.root / "packages" / "foo" / "pyproject.toml"
    member_pyproject.write_text(
        '[project]\nname = "foo"\nversion = "0.2.0"\ndependencies = []\n',
        encoding="utf-8",
    )
    fresh_project = core.create_project(
        workspace_project.root, global_config=workspace_project.global_config.config_file.as_posix()
    )

    assert not fresh_project.is_lockfile_hash_match()
    assert not fresh_project.is_lockfile_fresh()
    result = pdm(["lock", "--check"], obj=fresh_project)
    assert result.exit_code == 1


@pytest.mark.usefixtures("repository")
@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
@pytest.mark.parametrize("dependency_kind", ["direct-url", "editable-url"])
def test_lock_inputs_track_local_directory_dependencies(project, lock_format, dependency_kind):
    project.project_config["lock.format"] = lock_format
    local_path = project.root / "local" / "foo"
    local_path.mkdir(parents=True)
    local_pyproject = local_path / "pyproject.toml"
    local_pyproject.write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = []\n',
        encoding="utf-8",
    )
    dependency = "file:///${PROJECT_ROOT}/local/foo"
    if dependency_kind == "direct-url":
        project.add_dependencies([f"foo @ {dependency}"])
    else:
        project.add_dependencies([f"-e {dependency}"], to_group="dev", dev=True)
    enable_lock_inputs(project)
    actions.do_lock(project)
    project.lockfile.reload()

    assert project.is_lockfile_fresh()
    local_pyproject.write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = ["requests"]\n',
        encoding="utf-8",
    )

    assert not project.is_lockfile_hash_match()
    assert not project.is_lockfile_fresh()


@pytest.mark.usefixtures("repository")
@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_check_validates_local_project_specifiers(project, lock_format):
    project.project_config["lock.format"] = lock_format
    local_path = project.root / "local" / "foo"
    local_path.mkdir(parents=True)
    local_pyproject = local_path / "pyproject.toml"
    local_pyproject.write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = ["requests"]\n',
        encoding="utf-8",
    )
    project.add_dependencies(["foo @ file:///${PROJECT_ROOT}/local/foo"])
    enable_lock_inputs(project)
    actions.do_lock(project)
    project.lockfile.reload()

    local_pyproject.write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = ["requests>=2"]\n',
        encoding="utf-8",
    )
    assert project.is_lockfile_fresh()

    local_pyproject.write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = ["requests<2"]\n',
        encoding="utf-8",
    )
    assert not project.is_lockfile_fresh()


def test_lock_inputs_track_relative_editable_directory(project):
    local_path = project.root / "local" / "foo"
    local_path.mkdir(parents=True)
    local_pyproject = local_path / "pyproject.toml"
    local_pyproject.write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = []\n',
        encoding="utf-8",
    )
    project.pyproject.settings["dev-dependencies"] = {"dev": ["-e ./local/foo"]}
    project.pyproject.write(show_message=False)
    previous_inputs = project.lock_inputs()

    local_pyproject.write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = ["requests"]\n',
        encoding="utf-8",
    )

    assert project.lock_inputs() != previous_inputs


def test_dynamic_local_directory_is_volatile(project):
    local_path = project.root / "local" / "foo"
    local_path.mkdir(parents=True)
    local_path.joinpath("pyproject.toml").write_text(
        '[project]\nname = "foo"\ndynamic = ["version", "dependencies"]\n',
        encoding="utf-8",
    )
    project.add_dependencies(["foo @ file:///${PROJECT_ROOT}/local/foo"])
    lock_inputs = project.lock_inputs()

    assert not lock_inputs_match(project, lock_inputs)


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_inputs_track_local_file_content(project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    local_file = project.root / "demo-0.0.1-py2.py3-none-any.whl"
    local_file.write_bytes((FIXTURES / "artifacts" / "demo-0.0.1-py2.py3-none-any.whl").read_bytes())
    project.add_dependencies(["demo @ file:///${PROJECT_ROOT}/demo-0.0.1-py2.py3-none-any.whl"])
    enable_lock_inputs(project)
    actions.do_lock(project)
    project.lockfile.reload()

    assert project.is_lockfile_fresh()
    with local_file.open("ab") as file_handler:
        file_handler.write(b"changed")

    assert not project.is_lockfile_hash_match()
    assert not project.is_lockfile_fresh()


@pytest.mark.usefixtures("repository")
@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_dependencies(project, lock_format):
    project.add_dependencies(["requests"])
    project.project_config["lock.format"] = lock_format
    actions.do_lock(project)
    assert project.lockfile.exists()
    assert project.lockfile._path.name == "pdm.lock" if lock_format == "pdm" else "pylock.toml"
    locked = project.get_locked_repository().candidates
    for package in ("requests", "idna", "chardet", "certifi"):
        assert package in locked


@pytest.mark.parametrize("args", [("-S", "static_urls"), ("--static-urls",)])
def test_lock_refresh(pdm, project, repository, args, core, mocker):
    project.add_dependencies(["requests"])
    result = pdm(["lock"], obj=project)
    assert result.exit_code == 0
    assert project.is_lockfile_hash_match()
    package = next(p for p in project.lockfile["package"] if p["name"] == "requests")
    assert not package.get("files")
    project.add_dependencies(["requests>=2.0"])
    url_hashes = {
        "http://example.com/requests-2.19.1-py3-none-any.whl": "sha256:abcdef123456",
        "http://example2.com/requests-2.19.1-py3-none-AMD64.whl": "sha256:abcdef123456",
        "http://example1.com/requests-2.19.1-py3-none-any.whl": "sha256:abcdef123456",
    }
    mocker.patch.object(
        core.repository_class,
        "get_hashes",
        side_effect=(
            lambda c: (
                [{"url": url, "file": Link(url).filename, "hash": hash} for url, hash in url_hashes.items()]
                if c.identify() == "requests"
                else []
            )
        ),
    )
    assert not project.is_lockfile_hash_match()
    result = pdm(["lock", "--refresh", "-v"], obj=project)
    assert result.exit_code == 0
    package = next(p for p in project.lockfile["package"] if p["name"] == "requests")
    assert package["files"] == [
        {"file": "requests-2.19.1-py3-none-AMD64.whl", "hash": "sha256:abcdef123456"},
        {"file": "requests-2.19.1-py3-none-any.whl", "hash": "sha256:abcdef123456"},
    ]
    assert project.is_lockfile_hash_match()
    result = pdm(["lock", "--refresh", *args, "-v"], obj=project)
    assert result.exit_code == 0
    package = next(p for p in project.lockfile["package"] if p["name"] == "requests")
    assert package["files"] == [{"url": url, "hash": hash} for url, hash in sorted(url_hashes.items())]


def test_lock_refresh_keep_consistent(pdm, project, repository):
    project.add_dependencies(["requests"])
    result = pdm(["lock"], obj=project)
    assert result.exit_code == 0
    assert project.is_lockfile_hash_match()
    previous = project.lockfile._path.read_text()
    result = pdm(["lock", "--refresh"], obj=project)
    assert result.exit_code == 0
    assert project.lockfile._path.read_text() == previous


def test_pylock_refresh_preserves_dependency_group_markers(pdm, project, repository):
    project.project_config["lock.format"] = "pylock"
    project.add_dependencies(["requests"], to_group="dev", dev=True)
    result = pdm(["lock", "-G", "dev"], obj=project)
    assert result.exit_code == 0
    package = next(p for p in project.lockfile["packages"] if p["name"] == "requests")
    assert package["marker"] == '"dev" in dependency_groups'

    result = pdm(["lock", "--refresh"], obj=project)
    assert result.exit_code == 0
    package = next(p for p in project.lockfile["packages"] if p["name"] == "requests")
    assert package["marker"] == '"dev" in dependency_groups'


def test_lock_check_no_change_success(pdm, project, repository):
    project.add_dependencies(["requests"])
    result = pdm(["lock"], obj=project)
    assert result.exit_code == 0
    assert project.is_lockfile_hash_match()

    result = pdm(["lock", "--check"], obj=project)
    assert result.exit_code == 0


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_check_uses_canonical_inputs(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.add_dependencies(["requests", "pytz"])
    enable_lock_inputs(project)
    pdm(["lock"], obj=project, strict=True)

    assert project.lockfile.lock_inputs == project.lock_inputs()
    assert project.lockfile.lock_inputs_state is LockInputsState.SUPPORTED
    if lock_format == "pdm":
        assert project.lockfile.file_version == str(project.lockfile.spec_version)
        assert "content_hash" not in project.lockfile._data["metadata"]
    else:
        assert "hashes" not in project.lockfile._data["tool"]["pdm"]
        assert "lock_inputs_version" not in project.lockfile._path.read_text(encoding="utf-8")
    project.pyproject.metadata["dependencies"] = ["pytz", "requests"]
    project.pyproject.write(show_message=False)
    project.lockfile.reload()

    assert not project.is_lockfile_hash_match()
    assert project.is_lockfile_fresh()
    result = pdm(["lock", "--check"], obj=project)
    assert result.exit_code == 0


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_inputs_omit_named_specifiers(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.add_dependencies(["requests>=2"])
    enable_lock_inputs(project)

    pdm(["lock"], obj=project, strict=True)

    assert project.lockfile.lock_inputs["project"]["groups"]["default"] == ["requests"]


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_check_validates_only_locked_groups(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.pyproject.metadata["optional-dependencies"] = {"http": ["requests"]}
    enable_lock_inputs(project)

    pdm(["lock"], obj=project, strict=True)

    assert project.lockfile.groups == ["default"]
    assert "requests" not in project.get_locked_repository().candidates
    assert project.is_lockfile_fresh()
    result = pdm(["lock", "--check"], obj=project)
    assert result.exit_code == 0


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_check_allows_compatible_specifier_change(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.add_dependencies(["requests"])
    enable_lock_inputs(project)
    pdm(["lock"], obj=project, strict=True)

    project.add_dependencies(["requests>=2"])
    project.lockfile.reload()

    assert not project.is_lockfile_hash_match()
    assert project.is_lockfile_fresh()
    result = pdm(["lock", "--check"], obj=project)
    assert result.exit_code == 0


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_check_rejects_incompatible_specifier_change(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.add_dependencies(["requests"])
    enable_lock_inputs(project)
    pdm(["lock"], obj=project, strict=True)

    project.add_dependencies(["requests<2"])
    project.lockfile.reload()

    assert not project.is_lockfile_fresh()
    result = pdm(["lock", "--check"], obj=project)
    assert result.exit_code == 1


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_check_always_validates_current_specifier(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.add_dependencies(["requests>=2"])
    enable_lock_inputs(project)
    pdm(["lock"], obj=project, strict=True)

    packages = project.lockfile._data["package"] if lock_format == "pdm" else project.lockfile._data["packages"]
    package = next(package for package in packages if package["name"] == "requests")
    package["version"] = "1.0.0"
    project.lockfile.write(show_message=False)
    project.lockfile.reload()

    assert project.lockfile.lock_inputs == project.lock_inputs()
    assert not project.is_lockfile_fresh()


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_check_uses_effective_override_specifier(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.pyproject.settings["resolution"] = {
        "lock_inputs": True,
        "overrides": {"requests": "2.19.1"},
    }
    project.add_dependencies(["requests>=3"])

    pdm(["lock"], obj=project, strict=True)

    assert project.is_lockfile_fresh()


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
@pytest.mark.parametrize("dependency_kind", ["direct", "override"])
@pytest.mark.parametrize("corruption", ["missing", "source"])
def test_lock_check_validates_effective_file_source(
    pdm,
    project,
    repository,
    lock_format,
    dependency_kind,
    corruption,
):
    project.project_config["lock.format"] = lock_format
    url = "http://fixtures.test/artifacts/demo-0.0.1-py2.py3-none-any.whl"
    if dependency_kind == "direct":
        project.add_dependencies([f"demo @ {url}"])
    else:
        project.pyproject.settings["resolution"] = {"overrides": {"demo": url}}
        project.add_dependencies(["demo"])
    enable_lock_inputs(project)
    pdm(["lock"], obj=project, strict=True)

    packages = project.lockfile._data["package"] if lock_format == "pdm" else project.lockfile._data["packages"]
    package = next(package for package in packages if package["name"] == "demo")
    if corruption == "missing":
        packages.remove(package)
    elif lock_format == "pdm":
        package["url"] = f"{url}?tampered=1"
    else:
        package["archive"]["url"] = f"{url}?tampered=1"
    project.lockfile.write(show_message=False)
    project.lockfile.reload()

    assert project.lockfile.lock_inputs == project.lock_inputs()
    assert not project.is_lockfile_fresh()


@pytest.mark.usefixtures("vcs")
@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_check_validates_vcs_source(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.add_dependencies(["demo @ git+https://github.com/test-root/demo.git@1234567890abcdef"])
    enable_lock_inputs(project)
    pdm(["lock"], obj=project, strict=True)

    assert project.is_lockfile_fresh()
    packages = project.lockfile._data["package"] if lock_format == "pdm" else project.lockfile._data["packages"]
    package = next(package for package in packages if package["name"] == "demo")
    if lock_format == "pdm":
        package["git"] = "https://github.com/test-root/demo-tampered.git"
    else:
        package["vcs"]["url"] = "https://github.com/test-root/demo-tampered.git"
    project.lockfile.write(show_message=False)
    project.lockfile.reload()

    assert project.lockfile.lock_inputs == project.lock_inputs()
    assert not project.is_lockfile_fresh()


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_check_matches_specifiers_by_marker_context(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.pyproject.metadata["dependencies"] = [
        'django<2; sys_platform == "win32"',
        'django>=2; sys_platform != "win32"',
    ]
    project.pyproject.settings.setdefault("resolution", {})["lock_inputs"] = True
    project.pyproject.write(show_message=False)

    pdm(["lock", "--platform", "windows"], obj=project, strict=True)
    pdm(["lock", "--platform", "linux", "--append"], obj=project, strict=True)

    assert project.is_lockfile_fresh()
    project.pyproject.metadata["dependencies"][1] = 'django>=999; sys_platform != "win32"'
    project.pyproject.write(show_message=False)
    project.lockfile.reload()
    assert not project.is_lockfile_fresh()


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_check_requires_candidate_coverage_for_all_targets(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    repository.add_candidate("forked", "1.0", "<3.11")
    repository.add_candidate("forked", "2.0", ">=3.11")
    project.add_dependencies(["forked"])
    enable_lock_inputs(project)

    pdm(["lock", "--python", ">=3.10,<3.11"], obj=project, strict=True)
    pdm(["lock", "--python", ">=3.12,<3.13", "--append"], obj=project, strict=True)

    packages = project.lockfile._data["package"] if lock_format == "pdm" else project.lockfile._data["packages"]
    candidates = [package for package in packages if package["name"] == "forked"]
    assert len(candidates) == 2
    assert project.is_lockfile_fresh()

    packages.remove(candidates[0])
    project.lockfile.write(show_message=False)
    project.lockfile.reload()

    assert project.lockfile.lock_inputs == project.lock_inputs()
    assert not project.is_lockfile_fresh()


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_check_skips_excluded_named_dependency(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.pyproject.settings["resolution"] = {"excludes": ["requests"], "lock_inputs": True}
    local_path = project.root / "local" / "foo"
    local_path.mkdir(parents=True)
    local_path.joinpath("pyproject.toml").write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = ["requests>=999"]\n',
        encoding="utf-8",
    )
    project.add_dependencies(["foo @ file:///${PROJECT_ROOT}/local/foo"])

    pdm(["lock"], obj=project, strict=True)

    assert project.is_lockfile_fresh()


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
@pytest.mark.parametrize("lock_inputs_setting", [None, False])
def test_lock_check_falls_back_to_legacy_hash(pdm, project, repository, lock_format, lock_inputs_setting):
    project.project_config["lock.format"] = lock_format
    project.add_dependencies(["requests"])
    if lock_inputs_setting is not None:
        project.pyproject.open_for_write()
        project.pyproject.settings.setdefault("resolution", {})["lock_inputs"] = lock_inputs_setting
        project.pyproject.write(show_message=False)
    pdm(["lock"], obj=project, strict=True)

    assert project.lockfile.lock_inputs is None
    assert project.lockfile.lock_inputs_state is LockInputsState.LEGACY
    assert project.is_lockfile_fresh()


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_inputs_setting_requires_regeneration(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.add_dependencies(["requests"])
    pdm(["lock"], obj=project, strict=True)

    enable_lock_inputs(project)
    project.lockfile.reload()

    assert not project.is_lockfile_hash_match()
    assert not project.is_lockfile_fresh()

    pdm(["lock"], obj=project, strict=True)

    assert project.lockfile.lock_inputs_state is LockInputsState.SUPPORTED
    assert project.lockfile.lock_inputs is not None
    assert "lock_inputs" not in project.lockfile.lock_inputs["project"]["resolution"]
    assert not project.is_lockfile_hash_match()
    assert project.is_lockfile_fresh()


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_existing_lock_inputs_remain_enabled(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.add_dependencies(["requests"])
    enable_lock_inputs(project)
    pdm(["lock"], obj=project, strict=True)

    project.pyproject.settings["resolution"].pop("lock_inputs")
    project.pyproject.write(show_message=False)
    project.lockfile.reload()

    assert not project.is_lockfile_hash_match()
    assert project.lock_inputs_enabled()
    assert project.is_lockfile_fresh()

    pdm(["lock"], obj=project, strict=True)

    assert project.lockfile.lock_inputs_state is LockInputsState.SUPPORTED
    assert project.lockfile.lock_inputs is not None
    assert not project.is_lockfile_hash_match()
    assert project.is_lockfile_fresh()


@pytest.mark.parametrize("lock_version", ["", "4.5.0", "4.5.1", "4.5.2", "4.6.0"])
def test_pdmlock_uses_present_inputs_regardless_of_version(pdm, project, repository, lock_version):
    project.add_dependencies(["requests"])
    enable_lock_inputs(project)
    pdm(["lock"], obj=project, strict=True)
    project.lockfile._data["metadata"]["lock_version"] = lock_version
    project.pyproject.settings["resolution"].pop("lock_inputs")
    project.pyproject.write(show_message=False)

    assert project.lockfile.lock_inputs is not None
    assert project.lockfile.lock_inputs_state is LockInputsState.SUPPORTED
    assert not project.is_lockfile_hash_match()
    assert project.is_lockfile_fresh()


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_check_rejects_malformed_inputs(pdm, project, repository, lock_format):
    project.project_config["lock.format"] = lock_format
    project.add_dependencies(["requests"])
    enable_lock_inputs(project)
    pdm(["lock"], obj=project, strict=True)
    project.pyproject.settings["resolution"].pop("lock_inputs")
    project.pyproject.write(show_message=False)
    pdm(["lock"], obj=project, strict=True)
    if lock_format == "pdm":
        lock_inputs = project.lockfile._data["metadata"]["lock_inputs"]
    else:
        lock_inputs = project.lockfile._data["tool"]["pdm"]["lock_inputs"]
    lock_inputs["project"]["groups"]["default"] = [42]

    assert not project.is_lockfile_fresh()


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
@pytest.mark.parametrize("malformed_inputs", [{}, "invalid", datetime(2026, 1, 1, tzinfo=timezone.utc)])
def test_lock_check_does_not_fallback_for_invalid_lock_inputs(pdm, project, repository, lock_format, malformed_inputs):
    project.project_config["lock.format"] = lock_format
    project.add_dependencies(["requests"])
    enable_lock_inputs(project)
    pdm(["lock"], obj=project, strict=True)
    project.pyproject.settings["resolution"].pop("lock_inputs")
    project.pyproject.write(show_message=False)
    pdm(["lock"], obj=project, strict=True)
    if lock_format == "pdm":
        project.lockfile._data["metadata"]["lock_inputs"] = malformed_inputs
    else:
        project.lockfile._data["tool"]["pdm"]["lock_inputs"] = malformed_inputs

    assert not project.is_lockfile_hash_match()
    assert not project.is_lockfile_fresh()


def test_lock_check_rejects_nested_temporal_value(project):
    timestamp = "2026-01-01T00:00:00"
    project.pyproject.settings["source"] = [{"name": timestamp, "url": "https://example.org/simple"}]
    lock_inputs = project.lock_inputs()
    lock_inputs["project"]["sources"][0]["name"] = datetime.fromisoformat(timestamp)

    assert not lock_inputs_match(project, lock_inputs)


def test_lock_inputs_redact_credentials(project):
    project.pyproject.settings["source"] = [
        {
            "name": "private",
            "url": "https://url-user:url-password@example.org/source-path-secret?token=source-query-secret",
            "username": "config-user",
            "password": "config-password",
        }
    ]
    project.pyproject.metadata["dependencies"] = [
        (
            "demo @ https://dependency-user:dependency-password@example.org/dependency-path-secret/demo.whl"
            "?token=dependency-query-secret"
        )
    ]
    project.pyproject.settings["resolution"] = {
        "overrides": {
            "demo": "https://example.org/override-path-secret/demo.whl?token=override-query-secret",
        }
    }

    lock_inputs = repr(project.lock_inputs())

    assert "url-user" not in lock_inputs
    assert "url-password" not in lock_inputs
    assert "config-user" not in lock_inputs
    assert "config-password" not in lock_inputs
    assert "dependency-user" not in lock_inputs
    assert "dependency-password" not in lock_inputs
    assert "source-path-secret" not in lock_inputs
    assert "source-query-secret" not in lock_inputs
    assert "dependency-path-secret" not in lock_inputs
    assert "dependency-query-secret" not in lock_inputs
    assert "override-path-secret" not in lock_inputs
    assert "override-query-secret" not in lock_inputs
    assert "sha256:" in lock_inputs


@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_write_lockfile_does_not_add_lock_inputs(project, lock_format):
    project.project_config["lock.format"] = lock_format
    project.lockfile.open_for_write()

    project.write_lockfile(show_message=False)

    assert project.lockfile.lock_inputs is None


def test_lock_check_change_fails(pdm, project, repository):
    project.add_dependencies(["requests"])
    result = pdm(["lock"], obj=project)
    assert result.exit_code == 0
    assert project.is_lockfile_hash_match()

    project.add_dependencies(["pyyaml"])
    result = pdm(["lock", "--check"], obj=project)
    assert result.exit_code == 1


@pytest.mark.usefixtures("repository")
def test_innovations_with_specified_lockfile(pdm, project, working_set):
    project.add_dependencies(["requests"])
    lockfile = str(project.root / "mylock.lock")
    pdm(["lock", "--lockfile", lockfile], strict=True, obj=project)
    assert project.lockfile._path == project.root / "mylock.lock"
    assert project.is_lockfile_hash_match()
    locked = project.get_locked_repository().candidates
    assert "requests" in locked
    pdm(["sync", "--lockfile", lockfile], strict=True, obj=project)
    assert "requests" in working_set


@pytest.mark.usefixtures("repository", "vcs")
def test_skip_editable_dependencies_in_metadata(project, capsys):
    project.pyproject.metadata["dependencies"] = [
        "-e git+https://github.com/test-root/demo.git@1234567890abcdef#egg=demo"
    ]
    actions.do_lock(project)
    _, err = capsys.readouterr()
    assert "WARNING: Skipping editable dependency" in err
    assert not project.get_locked_repository().candidates


@pytest.mark.usefixtures("repository")
def test_lock_selected_groups(project, pdm):
    project.add_dependencies(["requests"], to_group="http")
    project.add_dependencies(["pytz"])
    pdm(["lock", "-G", "http", "--no-default"], obj=project, strict=True)
    assert project.lockfile.groups == ["http"]
    assert "requests" in project.get_locked_repository().candidates
    assert "pytz" not in project.get_locked_repository().candidates


@pytest.mark.usefixtures("repository")
@pytest.mark.parametrize("to_dev", [True, False])
def test_lock_self_referencing_dev_groups(project, pdm, to_dev):
    name = project.name
    project.add_dependencies(["requests"], to_group="http", dev=to_dev)
    project.add_dependencies(
        [parse_requirement("pytz"), parse_requirement(f"{name}[http]")],
        to_group="dev",
        dev=True,
    )
    pdm(["lock", "-G", "dev"], obj=project, strict=True)
    assert project.lockfile.groups == ["default", "dev", "http"]
    packages = project.lockfile["package"]
    pytz = next(p for p in packages if p["name"] == "pytz")
    assert pytz["groups"] == ["dev"]
    requests = next(p for p in packages if p["name"] == "requests")
    assert requests["groups"] == ["dev", "http"]
    idna = next(p for p in packages if p["name"] == "idna")
    assert idna["groups"] == ["dev", "http"]


@pytest.mark.usefixtures("repository")
def test_lock_self_referencing_optional_groups(project, pdm):
    name = project.name
    project.add_dependencies(["requests"], to_group="http")
    project.add_dependencies(
        [parse_requirement("pytz"), parse_requirement(f"{name}[http]")],
        to_group="all",
    )
    pdm(["lock", "-G", "all"], obj=project, strict=True)
    assert project.lockfile.groups == ["default", "all", "http"]
    packages = project.lockfile["package"]
    pytz = next(p for p in packages if p["name"] == "pytz")
    assert pytz["groups"] == ["all"]
    requests = next(p for p in packages if p["name"] == "requests")
    assert requests["groups"] == ["all", "http"]
    idna = next(p for p in packages if p["name"] == "idna")
    assert idna["groups"] == ["all", "http"]


@pytest.mark.usefixtures("repository")
def test_lock_include_groups_not_allowed(project, pdm):
    project.pyproject.metadata["optional-dependencies"] = {"http": ["requests"]}
    project.pyproject.dependency_groups.update({"dev": ["pytest", {"include-group": "http"}]})
    project.pyproject.write()
    result = pdm(["lock", "-G", "all"], obj=project)
    assert result.exit_code != 0
    assert "Missing group 'http' in `include-group`" in result.stderr


@pytest.mark.usefixtures("repository")
def test_lock_optional_referencing_dev_group_not_allowed(project, pdm):
    name = project.name
    project.pyproject.metadata["optional-dependencies"] = {"http": ["requests", f"{name}[dev]"]}
    project.pyproject.dependency_groups.update({"dev": ["pytest"]})
    project.pyproject.write()
    result = pdm(["lock", "-G", "http"], obj=project)
    assert result.exit_code != 0
    assert "Optional dependency group 'http' cannot include non-existing extras" in result.stderr


@pytest.mark.usefixtures("local_finder")
def test_lock_multiple_platform_wheels(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.7")
    project.add_dependencies(["pdm-hello"])
    pdm(["lock"], obj=project, strict=True)
    package = next(p for p in project.lockfile["package"] if p["name"] == "pdm-hello")
    file_hashes = package["files"]
    assert len(file_hashes) == 2


@pytest.mark.usefixtures("local_finder")
@pytest.mark.parametrize("platform", ["linux", "macos", "windows"])
def test_lock_specific_platform_wheels(project, pdm, platform):
    project.environment.python_requires = PySpecSet(">=3.7")
    project.add_dependencies(["pdm-hello"])
    pdm(["lock", "--platform", platform], obj=project, strict=True)
    assert FLAG_CROSS_PLATFORM not in project.lockfile.strategy
    package = next(p for p in project.lockfile["package"] if p["name"] == "pdm-hello")
    file_hashes = package["files"]
    wheels_num = 2 if platform == "windows" else 1
    assert len(file_hashes) == wheels_num


def test_parse_lock_strategy_group_options(core):
    core.init_parser()
    parser = core.parser

    ns = parser.parse_args(["lock", "-S", "no_cross_platform"])
    assert ns.strategy_change == ["no_cross_platform"]
    ns = parser.parse_args(["lock", "-S", "no_cross_platform", "--static-urls"])
    assert ns.strategy_change == ["no_cross_platform", "static_urls"]
    ns = parser.parse_args(["lock", "-S", "no_cross_platform,direct_minimal_versions"])
    assert ns.strategy_change == ["no_cross_platform", "direct_minimal_versions"]


def test_apply_lock_strategy_changes(project):
    assert project.lockfile.apply_strategy_change(["no_cross_platform", "static_urls"]) == {
        "inherit_metadata",
        "static_urls",
    }
    assert project.lockfile.apply_strategy_change(["no_static_urls"]) == {"inherit_metadata"}
    assert project.lockfile.apply_strategy_change(["no_inherit_metadata"]) == set()


@pytest.mark.parametrize("strategy", [["abc"], ["no_abc", "static_urls"]])
def test_apply_lock_strategy_changes_invalid(project, strategy):
    with pytest.raises(PdmUsageError):
        project.lockfile.apply_strategy_change(strategy)


def test_lock_direct_minimal_versions(project, repository, pdm):
    project.add_dependencies(["django"])
    repository.add_candidate("pytz", "2019.6")
    pdm(["lock", "-S", "direct_minimal_versions"], obj=project, strict=True)
    assert project.lockfile.strategy == {"direct_minimal_versions", "inherit_metadata"}
    locked_repository = project.get_locked_repository()
    assert locked_repository.candidates["django"].version == "1.11.8"
    assert locked_repository.candidates["pytz"].version == "2019.6"


@pytest.mark.usefixtures("local_finder")
@pytest.mark.parametrize("args", [(), ("-S", "direct_minimal_versions")])
def test_lock_direct_minimal_versions_real(project, pdm, args):
    project.add_dependencies(["zipp"])
    pdm(["lock", *args], obj=project, strict=True)
    locked_candidate = project.get_locked_repository().candidates["zipp"]
    if args:
        assert locked_candidate.version == "3.6.0"
    else:
        assert locked_candidate.version == "3.7.0"


@pytest.mark.parametrize(
    "lock_version,expected",
    [
        ("4.1.0", Compatibility.BACKWARD),
        ("4.1.1", Compatibility.SAME),
        ("4.1.2", Compatibility.FORWARD),
        ("4.2", Compatibility.NONE),
        ("3.0", Compatibility.NONE),
        ("4.0.1", Compatibility.BACKWARD),
    ],
)
def test_lockfile_compatibility(project, monkeypatch, lock_version, expected, pdm):
    pdm(["lock"], obj=project, strict=True)
    monkeypatch.setattr("pdm.project.lockfile.PDMLock.spec_version", parse_version("4.1.1"))
    project.lockfile._data["metadata"]["lock_version"] = lock_version
    assert project.lockfile.compatibility() == expected
    result = pdm(["lock", "--check"], obj=project)
    assert result.exit_code == (1 if expected == Compatibility.NONE else 0)


def test_lock_default_inherit_metadata(project, pdm, mocker, working_set):
    project.add_dependencies(["requests"])
    pdm(["lock"], obj=project, strict=True)
    assert "inherit_metadata" in project.lockfile.strategy
    packages = project.lockfile["package"]
    assert all(package["groups"] == ["default"] for package in packages)

    resolver = mocker.patch.object(project, "get_resolver")
    pdm(["sync"], obj=project, strict=True)
    resolver.assert_not_called()
    for key in ("requests", "idna", "chardet", "urllib3"):
        assert key in working_set


def test_lock_inherit_metadata_strategy(project, pdm, mocker, working_set):
    project.add_dependencies(["requests"])
    pdm(["lock", "-S", "inherit_metadata"], obj=project, strict=True)
    assert "inherit_metadata" in project.lockfile.strategy
    packages = project.lockfile["package"]
    assert all(package["groups"] == ["default"] for package in packages)

    resolver = mocker.patch.object(project, "get_resolver")
    pdm(["sync"], obj=project, strict=True)
    resolver.assert_not_called()
    for key in ("requests", "idna", "chardet", "urllib3"):
        assert key in working_set


@pytest.mark.parametrize("source", ["cli", "pyproject", "config"])
def test_lock_exclude_newer(project, pdm, source):
    project.pyproject.metadata["requires-python"] = ">=3.9"
    project.project_config["pypi.url"] = "https://my.pypi.org/json"
    project.add_dependencies(["zipp"])
    pdm(["lock"], obj=project, strict=True, cleanup=False)
    assert project.get_locked_repository().candidates["zipp"].version == "3.7.0"

    cmd = ["lock", "--exclude-newer", "2024-01-01"]
    if source == "pyproject":
        project.pyproject.settings["resolution"] = {"exclude-newer": "2024-01-01"}
        cmd = ["lock"]
    elif source == "config":
        project.project_config["strategy.exclude-newer"] = "2024-01-01"
        cmd = ["lock"]

    pdm(cmd, obj=project, strict=True, cleanup=False)
    assert project.get_locked_repository().candidates["zipp"].version == "3.6.0"


@pytest.mark.parametrize("source", ["cli", "pyproject", "config"])
def test_lock_exclude_newer_accepts_relative_duration(project, pdm, monkeypatch, source):
    from datetime import datetime, timezone

    import pdm.utils as pdm_utils

    frozen_now = datetime(2024, 2, 9, 12, 0, tzinfo=timezone.utc)

    class FrozenDateTime(datetime):
        @classmethod
        def now(cls, tz=None):
            assert tz is timezone.utc
            return frozen_now

    monkeypatch.setattr(pdm_utils, "datetime", FrozenDateTime)

    cmd = ["lock", "--exclude-newer", "3w"]
    if source == "pyproject":
        project.pyproject.settings["resolution"] = {"exclude-newer": "3w"}
        cmd = ["lock"]
    elif source == "config":
        project.project_config["strategy.exclude-newer"] = "3w"
        cmd = ["lock"]

    project.pyproject.metadata["requires-python"] = ">=3.9"
    project.project_config["pypi.url"] = "https://my.pypi.org/json"
    project.add_dependencies(["zipp"])
    pdm(cmd, strict=True, obj=project, cleanup=False)
    assert project.get_locked_repository().candidates["zipp"].version == "3.6.0"


def test_pyproject_exclude_newer_overrides_config(project, pdm):
    project.pyproject.metadata["requires-python"] = ">=3.9"
    project.project_config["pypi.url"] = "https://my.pypi.org/json"
    project.project_config["strategy.exclude-newer"] = "2024-01-01"
    project.pyproject.settings["resolution"] = {"exclude-newer": "2025-01-01"}
    project.add_dependencies(["zipp"])

    pdm(["lock"], obj=project, strict=True, cleanup=False)
    assert project.get_locked_repository().candidates["zipp"].version == "3.7.0"


exclusion_cases = [
    pytest.param(("-G", ":all", "--without", "tz,ssl"), id="-G :all --without tz,ssl"),
    pytest.param(("-G", ":all", "--without", "tz", "--without", "ssl"), id="-G :all --without tz --without ssl"),
    pytest.param(("--with", ":all", "--without", "tz,ssl"), id="--with all --without tz,ssl"),
    pytest.param(("--with", ":all", "--without", "tz", "--without", "ssl"), id="--with all --without tz --without ssl"),
    pytest.param(("--without", "tz", "--without", "ssl"), id="--without tz --without ssl"),
    pytest.param(("--without", "tz,ssl"), id="--without tz,ssl"),
]


@pytest.mark.parametrize("args", exclusion_cases)
@pytest.mark.usefixtures("repository")
def test_lock_all_with_excluded_groups(project, pdm, args):
    project.add_dependencies(["urllib3"], "url")
    project.add_dependencies(["pytz"], "tz", True)
    project.add_dependencies(["pyopenssl"], "ssl")
    pdm(["lock", *args], obj=project, strict=True)
    assert "urllib3" in project.get_locked_repository().candidates
    assert "pytz" not in project.get_locked_repository().candidates
    assert "pyopenssl" not in project.get_locked_repository().candidates


@pytest.mark.parametrize(
    "args",
    [
        ("--append",),
        ("--python", "<3.6"),
        ("-S", "cross_platform", "--append", "--python", "3.10"),
        ("--platform", "linux", "--refresh"),
    ],
)
def test_forbidden_lock_target_options(project, pdm, args):
    result = pdm(["lock", *args], obj=project)
    assert result.exit_code != 0
    assert "PdmUsageError" in result.stderr


@pytest.mark.parametrize("nested", [False, True])
def test_lock_for_multiple_targets(project, pdm, repository, nested):
    deps = [
        'django<2; sys_platform == "win32"',
        'django>=2; sys_platform != "win32"',
    ]
    if nested:
        repository.add_candidate("foo", "0.1.0")
        repository.add_dependencies("foo", "0.1.0", deps)
        project.add_dependencies(["foo"])
    else:
        project.add_dependencies(deps)

    pdm(["lock", "--platform", "windows"], obj=project, strict=True)
    locked = project.get_locked_repository()
    candidates = locked.all_candidates
    assert len(candidates["django"]) == 1
    assert candidates["django"][0].version == "1.11.8"
    assert len(locked.targets) == 1
    pytz = candidates["pytz"][0]
    assert str(pytz.req.marker) == 'sys_platform == "win32"'

    result = pdm(["lock", "--platform", "windows", "--append"], obj=project, strict=True)
    assert "already exists, skip locking." in result.stdout

    pdm(["lock", "--platform", "linux", "--append"], obj=project, strict=True)
    locked = project.get_locked_repository()
    candidates = locked.all_candidates
    assert len(locked.targets) == 2
    assert sorted(c.version for c in candidates["django"]) == ["1.11.8", "2.2.9"]
    pytz = candidates["pytz"][0]
    assert not pytz.req.marker or pytz.req.marker.is_any()

    # not append but overwrite
    pdm(["lock", "--platform", "windows"], obj=project, strict=True)
    locked = project.get_locked_repository()
    candidates = locked.all_candidates
    assert len(candidates["django"]) == 1
    assert candidates["django"][0].version == "1.11.8"
    assert len(locked.targets) == 1
    pytz = candidates["pytz"][0]
    assert str(pytz.req.marker) == 'sys_platform == "win32"'


CONSTRAINT_FILE = str(FIXTURES / "constraints.txt")


@pytest.mark.usefixtures("repository")
@pytest.mark.parametrize("constraint", [CONSTRAINT_FILE, Path(CONSTRAINT_FILE).as_uri()])
def test_lock_with_override_file(project, pdm, constraint):
    project.add_dependencies(["requests"])
    pdm(["lock", "--override", constraint], obj=project, strict=True)
    candidates = project.get_locked_repository().candidates
    assert candidates["requests"].version == "2.20.0b1"
    assert candidates["urllib3"].version == "1.23b0"
    assert "django" not in candidates


def test_pylock_add_remove_strategy(project, pdm):
    project.project_config["lock.format"] = "pylock"
    pdm(["lock"], obj=project, strict=True)
    assert project.lockfile.strategy == {"inherit_metadata", "static_urls"}
    pdm(["lock", "-S", "static_urls"], obj=project, strict=True)
    pdm(["lock", "-S", "direct_minimal_versions"], obj=project, strict=True)
    assert project.lockfile.strategy == {"inherit_metadata", "static_urls", "direct_minimal_versions"}

    result = pdm(["lock", "-S", "no_static_urls"], obj=project)
    assert result.exit_code != 0
    result = pdm(["lock", "-S", "no_inherit_metadata"], obj=project)
    assert result.exit_code != 0


@pytest.mark.usefixtures("repository")
def test_lock_with_invalid_python_requirement(project, pdm):
    project.add_dependencies(["requests", "python>=3.6"])
    result = pdm(["lock", "-v"], obj=project, strict=True)
    assert "requests" in project.get_locked_repository().candidates
    assert "python" not in project.get_locked_repository().candidates
    assert "The 'python' requirement is not necessary and will be ignored." in result.stderr



from pathlib import Path

from argcomplete.shell_integration import shellcode

COMPLETIONS = Path(__file__).parent.parent / "src/pdm/cli/completions"


def main():
    for shell, suffix, title in (
        ("bash", "bash", "BASH"),
        ("zsh", "zsh", "ZSH"),
        ("fish", "fish", "FISH"),
        ("powershell", "ps1", "Powershell"),
    ):
        header = f"# {title} completion script for pdm\n"
        COMPLETIONS.joinpath(f"pdm.{suffix}").write_text(header + shellcode(["pdm"], shell=shell))


if __name__ == "__main__":
    main()



import pytest

from pdm.exceptions import InvalidPyVersion
from pdm.models.specifiers import PySpecSet, _convert_spec, _fix_py4k, get_specifier
from pdm.models.versions import Version


@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.parametrize(
    "original,normalized",
    [
        (">=3.6", ">=3.6"),
        ("<3.8", "<3.8"),
        ("~=2.7.0", "~=2.7.0"),
        ("", ""),
        (">=3.6,<3.8", "<3.8,>=3.6"),
        (">3.6", ">3.6"),
        ("<=3.7", "<=3.7"),
        (">=3.4.*", ">=3.4.0"),
        (">3.4.*", ">=3.4.0"),
        ("<=3.4.*", "<3.4.0"),
        ("<3.4.*", "<3.4.0"),
        (">=3.0+g1234", ">=3.0"),
        ("<3.0+g1234", "<3.0"),
        ("<3.10.0a6", "<3.10.0a6"),
        ("<3.10.2a3", "<3.10.2a3"),
    ],
)
def test_normalize_pyspec(original, normalized):
    spec = PySpecSet(original)
    assert str(spec) == normalized


@pytest.mark.parametrize(
    "left,right,result",
    [
        (">=3.6", ">=3.0", ">=3.6"),
        (">=3.6", "<3.8", "<3.8,>=3.6"),
        ("", ">=3.6", ">=3.6"),
        (">=3.6", "<3.2", "<empty>"),
        (">=2.7,!=3.0.*", "!=3.1.*", "!=3.0.*,!=3.1.*,>=2.7"),
        (">=3.11.0a2", "<3.11.0b", ">=3.11.0a2,<3.11.0b0"),
        ("<3.11.0a2", ">3.11.0b", "<empty>"),
    ],
)
def test_pyspec_and_op(left, right, result):
    left = PySpecSet(left)
    right = PySpecSet(right)
    assert left & right == PySpecSet(result)


@pytest.mark.parametrize(
    "left,right,result",
    [
        (">=3.6", ">=3.0", ">=3.0"),
        ("", ">=3.6", ""),
        (">=3.6", "<3.7", ""),
        (">=3.6,<3.8", ">=3.4,<3.7", "<3.8,>=3.4"),
        ("~=2.7", ">=3.6", "!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*,>=2.7"),
        ("<2.7.15", ">=3.0", "!=2.7.15,!=2.7.16,!=2.7.17,!=2.7.18"),
        (">3.11.0a2", ">3.11.0b", ">3.11.0a2"),
    ],
)
def test_pyspec_or_op(left, right, result):
    left = PySpecSet(left)
    right = PySpecSet(right)
    assert str(left | right) == result


def test_impossible_pyspec():
    spec = PySpecSet(">=3.6,<3.4")
    a = PySpecSet(">=2.7")
    assert spec.is_empty()
    assert (spec & a).is_empty()
    assert spec | a == a


@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.parametrize(
    "left,right",
    [
        ("~=2.7", ">=2.7"),
        (">=3.6", ""),
        (">=3.7", ">=3.6,<4.0"),
        (">=2.7,<3.0", ">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*"),
        (">=3.6", ">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*"),
        (
            ">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*",
            ">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*",
        ),
        (">=3.11.*", ">=3.11.0rc"),
    ],
)
def test_pyspec_is_subset_superset(left, right):
    left = PySpecSet(left)
    right = PySpecSet(right)
    assert left.is_subset(right), f"{left}, {right}"
    assert right.is_superset(left), f"{left}, {right}"


@pytest.mark.parametrize(
    "left,right",
    [
        ("~=2.7", ">=2.6,<2.7.15"),
        (">=3.7", ">=3.6,<3.9"),
        (">=3.7,<3.6", "==2.7"),
        (">=3.0,!=3.4.*", ">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*"),
        (">=3.11.0", "<3.11.0a"),
    ],
)
def test_pyspec_isnot_subset_superset(left, right):
    left = PySpecSet(left)
    right = PySpecSet(right)
    assert not left.is_subset(right), f"{left}, {right}"
    assert not left.is_superset(right), f"{left}, {right}"


def test_get_specifier_accepts_empty_values():
    assert str(get_specifier(None)) == ""
    assert str(get_specifier("*")) == ""
    assert str(get_specifier(">=3.9")) == ">=3.9"


def test_pyspec_rejects_invalid_version():
    with pytest.raises(InvalidPyVersion, match="Invalid specifier"):
        PySpecSet("not-a-specifier")


def test_pyspec_protocol_operations():
    spec = PySpecSet(">=3.9")
    logic = PySpecSet("<4")._logic

    assert spec.__eq__(object()) is NotImplemented
    assert spec & logic == PySpecSet(">=3.9,<4")
    assert spec | logic == PySpecSet()
    assert spec.__and__(object()) is NotImplemented
    assert spec.__or__(object()) is NotImplemented
    assert hash(spec) == hash(spec._logic)
    assert repr(spec) == "<PySpecSet >=3.9>"


@pytest.mark.parametrize(
    ("lower", "upper"),
    [
        ("3.9.0", "3.9.0"),
        ("2.7.0", "3.1.0"),
        ("3.9.0", "3.10.2"),
        ("3.9.1", "3.11.0"),
        ("3.9.1", "3.9.3"),
    ],
)
def test_populate_version_range_boundaries(lower, upper):
    assert list(PySpecSet._populate_version_range(Version(lower), Version(upper))) is not None


def test_superset_and_subset_shortcuts():
    empty = PySpecSet("<empty>")
    any_spec = PySpecSet()
    constrained = PySpecSet(">=3.9")

    assert not empty.is_superset(constrained)
    assert not empty.is_subset(constrained)
    assert any_spec.is_superset(">=3.9")
    assert constrained.is_subset("")


def test_as_marker_string_variants():
    assert PySpecSet().as_marker_string() == ""
    with pytest.raises(InvalidPyVersion, match="Impossible specifier"):
        PySpecSet("<empty>").as_marker_string()

    marker = PySpecSet(">=3.9,<4,!=3.10.*,!=3.11.1").as_marker_string()
    assert "python_version>='3.9'" in marker
    assert "3.10.0" in marker
    assert "3.11.1" in marker


def test_convert_union_specifier_and_fix_py4k():
    union = PySpecSet("<3.9")._logic | PySpecSet(">=3.11,<4")._logic

    assert " or " in _convert_spec(union)
    assert _fix_py4k(PySpecSet("<4.0")._logic).is_any()
    assert _fix_py4k(PySpecSet(">=3.9,<3.10")._logic) == PySpecSet(">=3.9,<3.10")._logic



"""Tests for the completion command"""

import io
import os
import shutil
import subprocess
import sys
from pathlib import Path
from unittest.mock import patch

import pytest
from argcomplete import CompletionFinder
from argcomplete.completers import SuppressCompleter
from argcomplete.shell_integration import shellcode

from pdm.cli.completions import configure_parser
from pdm.core import Core


class TestCompletionFinder(CompletionFinder):
    __test__ = False

    def _init_debug_stream(self):
        pass


def completion_values(*words):
    core = Core()
    configure_parser(core)
    finder = TestCompletionFinder(
        core.parser,
        always_complete_options=False,
        default_completer=SuppressCompleter(),
    )
    line = f"pdm {' '.join(words)}"
    output = io.StringIO()

    class CompletionFinished(Exception):
        pass

    def exit_method(_status):
        raise CompletionFinished

    environment = {
        "COMP_LINE": line,
        "COMP_POINT": str(len(line)),
        "_ARGCOMPLETE": "1",
        "_ARGCOMPLETE_IFS": "\n",
        "_ARGCOMPLETE_SHELL": "bash",
    }
    with patch.dict(os.environ, environment), pytest.raises(CompletionFinished):
        finder(
            core.parser,
            always_complete_options=False,
            exit_method=exit_method,
            output_stream=output,
            default_completer=SuppressCompleter(),
        )
    return {value.rstrip() for value in output.getvalue().splitlines()}


def test_completion_bash(pdm):
    """Test completion for bash shell"""
    result = pdm(["completion", "bash"])
    assert result.exit_code == 0
    assert "BASH completion script for pdm" in result.output


@pytest.mark.skipif(
    sys.platform == "win32" or shutil.which("bash") is None,
    reason="requires a POSIX bash on PATH",
)
def test_completion_bash_runs_without_bash_completion_pkg(tmp_path):
    """Regression test for #3793.

    The generated bash completion script must work even when the
    ``bash-completion`` package is not loaded (e.g. Git Bash on Windows,
    minimal Linux containers). Before the fix, sourcing the script and
    triggering completion printed::

        bash: __ltrim_colon_completions: command not found

    in the middle of the candidate list.
    """
    bash_script = Path(__file__).resolve().parents[2] / "src" / "pdm" / "cli" / "completions" / "pdm.bash"
    project_root = bash_script.parents[4]
    driver = tmp_path / "drive.sh"
    driver.write_text(
        # Explicitly clear the bash-completion helpers so this test is
        # deterministic even when the host has bash-completion installed.
        "unset -f _get_comp_words_by_ref __ltrim_colon_completions 2>/dev/null\n"
        f"source {bash_script}\n"
        f'PATH="{project_root / ".venv" / "bin"}:$PATH"\n'
        "COMP_WORDS=(pdm '')\n"
        "COMP_CWORD=1\n"
        "COMP_LINE='pdm '\n"
        "COMP_POINT=4\n"
        "COMP_TYPE=9\n"
        # Reasonable default that includes ':' as a wordbreak char,
        # which is what makes __ltrim_colon_completions matter.
        "COMP_WORDBREAKS=$' \\t\\n\"\\'><=;|&(:'\n"
        "COMPREPLY=()\n"
        "_python_argcomplete pdm\n"
        # Surface candidates on stdout, errors on stderr.
        'printf "%s\\n" "${COMPREPLY[@]}"\n'
    )
    proc = subprocess.run(
        ["bash", str(driver)],
        capture_output=True,
        text=True,
        check=False,
    )
    assert proc.returncode == 0, proc.stderr
    # The exact symptoms of the bug:
    assert "command not found" not in proc.stderr, proc.stderr
    assert "_get_comp_words_by_ref" not in proc.stderr, proc.stderr
    assert "__ltrim_colon_completions" not in proc.stderr, proc.stderr
    # Sanity: candidates were produced.
    candidates = set(proc.stdout.split())
    assert {"add", "install", "venv"}.issubset(candidates), proc.stdout


def test_completion_zsh(pdm):
    """Test completion for zsh shell"""
    result = pdm(["completion", "zsh"])
    assert result.exit_code == 0
    assert "#compdef pdm" in result.output


def test_completion_fish(pdm):
    """Test completion for fish shell"""
    result = pdm(["completion", "fish"])
    assert result.exit_code == 0
    assert "FISH completion script for pdm" in result.output


def test_completion_powershell(pdm):
    """Test completion for powershell"""
    result = pdm(["completion", "powershell"])
    assert result.exit_code == 0
    assert "Powershell completion script for pdm" in result.output


def test_completion_pwsh(pdm):
    """Test completion for pwsh (PowerShell Core)"""
    result = pdm(["completion", "pwsh"])
    assert result.exit_code == 0
    assert "Powershell completion script for pdm" in result.output


@pytest.mark.parametrize(
    ("shell", "suffix", "title"),
    [
        ("bash", "bash", "BASH"),
        ("zsh", "zsh", "ZSH"),
        ("fish", "fish", "FISH"),
        ("powershell", "ps1", "Powershell"),
    ],
)
def test_completion_script_is_up_to_date(shell, suffix, title):
    script = Path(__file__).resolve().parents[2] / "src" / "pdm" / "cli" / "completions" / f"pdm.{suffix}"
    assert script.read_text() == f"# {title} completion script for pdm\n" + shellcode(["pdm"], shell=shell)


def test_completion_unsupported_shell(pdm):
    """Test completion with unsupported shell raises error"""
    result = pdm(["completion", "unsupported_shell"])
    assert result.exit_code != 0
    assert "Unsupported shell" in result.stderr


def test_completion_auto_detect(pdm, monkeypatch):
    """Test completion with auto-detected shell"""
    import shellingham

    monkeypatch.setattr(shellingham, "detect_shell", lambda: ("bash", "/bin/bash"))
    result = pdm(["completion"])
    assert result.exit_code == 0
    assert "BASH completion script for pdm" in result.output


def test_completion_auto_detect_unsupported(pdm, monkeypatch):
    """Test completion with auto-detected unsupported shell"""
    import shellingham

    monkeypatch.setattr(shellingham, "detect_shell", lambda: ("csh", "/bin/csh"))
    result = pdm(["completion"])
    assert result.exit_code != 0
    assert "Unsupported shell" in result.stderr


def test_completion_engine_completes_commands_and_options():
    assert {"add", "install", "venv"} <= completion_values("")
    assert {"clear", "info", "list", "remove"} <= completion_values("cache", "")
    assert {"create", "list", "remove"} <= completion_values("venv", "")
    assert {"--group", "--dry-run"} <= completion_values("add", "--")
    assert {"-G", "--group"} <= completion_values("add", "-")


def test_completion_engine_completes_static_values():
    assert completion_values("cache", "clear", "") == {"hashes", "http", "metadata", "packages", "wheels"}
    assert completion_values("export", "--format", "") == {"pylock", "requirements"}
    assert completion_values("export", "--format=r") == {"--format=requirements"}
    assert completion_values("completion", "") == {"bash", "fish", "powershell", "pwsh", "zsh"}


def test_completion_engine_completes_project_values(tmp_path, monkeypatch):
    tmp_path.joinpath("pyproject.toml").write_text(
        """
[project]
dependencies = ["httpx>=0.27"]

[project.optional-dependencies]
docs = ["sphinx"]

[dependency-groups]
test = ["pytest"]

[tool.pdm.scripts]
lint = "ruff check"
"""
    )
    monkeypatch.chdir(tmp_path)

    assert {"default", "docs", "test"} <= completion_values("install", "--group", "")
    assert completion_values("install", "--group", "docs,t") == {"docs,test"}
    assert {"httpx", "pytest", "sphinx"} <= completion_values("remove", "")
    assert completion_values("run", "") == {"lint"}
    assert "lint" in completion_values("")


@pytest.mark.skipif(sys.platform == "win32", reason="argcomplete file completion requires Bash")
def test_completion_engine_uses_file_directives():
    assert completion_values("--config", "")
    assert all(Path(value).is_dir() for value in completion_values("build", "--dest", ""))



from __future__ import annotations

import dataclasses

import pytest

from pdm.exceptions import BuildError
from pdm.installers.base import BaseSynchronizer, editables_candidate
from pdm.models.candidates import Candidate
from pdm.models.requirements import parse_requirement


def make_synchronizer(mocker):
    synchronizer = object.__new__(BaseSynchronizer)
    synchronizer.environment = mocker.Mock()
    synchronizer.environment.project.backend.expand_line.side_effect = lambda value: value
    synchronizer.reinstall = False
    synchronizer.no_editable = False
    synchronizer.install_self = False
    return synchronizer


@pytest.mark.parametrize("has_match", [False, True])
def test_editables_candidate(mocker, has_match):
    best = mocker.Mock() if has_match else None
    finder = mocker.MagicMock()
    finder.find_best_match.return_value.best = best
    environment = mocker.MagicMock()
    environment.get_finder.return_value.__enter__.return_value = finder
    from_installation_candidate = mocker.patch(
        "pdm.installers.base.Candidate.from_installation_candidate",
    )

    result = editables_candidate(environment)

    if has_match:
        assert result is from_installation_candidate.return_value
        from_installation_candidate.assert_called_once()
    else:
        assert result is None


@pytest.mark.parametrize(
    ("install_self", "has_editables", "requires", "build_error", "expected"),
    [
        (False, False, ["editables"], False, False),
        (True, True, ["editables"], False, False),
        (True, False, ["other"], False, False),
        (True, False, ["editables>=0.5"], False, True),
        (True, False, [], True, False),
    ],
)
def test_should_install_editables(mocker, install_self, has_editables, requires, build_error, expected):
    synchronizer = make_synchronizer(mocker)
    synchronizer.install_self = install_self
    synchronizer.requested_candidates = {"editables": mocker.Mock()} if has_editables else {}
    prepared = mocker.Mock()
    prepared.metadata.requires = requires
    candidate = mocker.Mock()
    candidate.prepare.return_value = prepared
    if build_error:
        candidate.prepare.side_effect = BuildError("failed")
    synchronizer.__dict__["self_candidate"] = candidate

    assert synchronizer.should_install_editables() is expected


def test_candidates_marks_requested_and_converts_editable(mocker):
    synchronizer = make_synchronizer(mocker)
    editable_req = dataclasses.replace(parse_requirement("demo"), editable=True)
    editable = Candidate(editable_req, name="demo", version="1.0")
    regular = Candidate(parse_requirement("regular"), name="regular", version="2.0")
    synchronizer.requested_candidates = {"demo": editable, "regular": regular}
    synchronizer.requirements = [parse_requirement("regular")]
    synchronizer.no_editable = ["demo"]

    result = synchronizer.candidates

    assert result["demo"].req.editable is False
    assert result["regular"].requested is True


def test_candidates_adds_editables_dependency(mocker):
    synchronizer = make_synchronizer(mocker)
    synchronizer.requested_candidates = {}
    synchronizer.requirements = []
    synchronizer.environment.project.all_dependencies = {}
    synchronizer.no_editable = False
    mocker.patch.object(synchronizer, "should_install_editables", return_value=True)
    candidate = mocker.patch("pdm.installers.base.editables_candidate").return_value

    assert synchronizer.candidates == {"editables": candidate}


def test_should_update_immediate_cases(mocker):
    synchronizer = make_synchronizer(mocker)
    dist = mocker.Mock()
    candidate = mocker.Mock()
    candidate.req.editable = False

    synchronizer.reinstall = True
    assert synchronizer._should_update(dist, candidate)
    synchronizer.reinstall = False
    candidate.req.editable = True
    assert synchronizer._should_update(dist, candidate)
    candidate.req.editable = False
    mocker.patch("pdm.installers.base.is_editable", return_value=True)
    assert not synchronizer._should_update(dist, candidate)
    synchronizer.no_editable = True
    assert synchronizer._should_update(dist, candidate)


@pytest.mark.parametrize(
    ("direct_url", "hashes", "expected"),
    [
        (None, [], False),
        ('{"url": "x"}', [], False),
        ('{"archive_info": {"hash": "sha256=abc"}}', [{"hash": "sha256:abc"}], False),
        ('{"archive_info": {"hash": "sha256=abc"}}', [{"hash": "sha256:def"}], True),
    ],
)
def test_should_update_archive_requirement(mocker, direct_url, hashes, expected):
    synchronizer = make_synchronizer(mocker)
    mocker.patch("pdm.installers.base.is_editable", return_value=False)
    dist = mocker.Mock()
    dist.read_text.return_value = direct_url
    installed_req = parse_requirement("demo @ https://example.org/demo.whl")
    mocker.patch("pdm.installers.base.Requirement.from_dist", return_value=installed_req)
    candidate = mocker.Mock()
    candidate.req.is_named = False
    candidate.req.editable = False
    candidate.link.url_without_fragment = "https://example.org/demo.whl"
    candidate.hashes = hashes

    assert synchronizer._should_update(dist, candidate) is expected


def test_should_update_changed_url_and_local_directory(mocker):
    synchronizer = make_synchronizer(mocker)
    mocker.patch("pdm.installers.base.is_editable", return_value=False)
    candidate = mocker.Mock()
    candidate.req.is_named = False
    candidate.req.editable = False
    candidate.link.url_without_fragment = "https://example.org/new.whl"
    installed_req = mocker.Mock()
    installed_req.is_local_dir = False
    installed_req.get_full_url.return_value = "https://example.org/old.whl"
    mocker.patch("pdm.installers.base.Requirement.from_dist", return_value=installed_req)
    mocker.patch("pdm.installers.base.FileRequirement", type(installed_req))

    assert synchronizer._should_update(mocker.Mock(), candidate)
    installed_req.is_local_dir = True
    assert synchronizer._should_update(mocker.Mock(), candidate)


def test_should_update_unknown_direct_requirement(mocker):
    synchronizer = make_synchronizer(mocker)
    mocker.patch("pdm.installers.base.is_editable", return_value=False)
    mocker.patch("pdm.installers.base.Requirement.from_dist", return_value=mocker.Mock())
    candidate = mocker.Mock()
    candidate.req.is_named = False
    candidate.req.editable = False

    assert synchronizer._should_update(mocker.Mock(), candidate)


def test_self_key_preserves_empty_project_name(mocker):
    synchronizer = make_synchronizer(mocker)
    synchronizer.install_self = True
    synchronizer.environment.project.name = ""

    assert synchronizer.self_key == ""


def test_synchronize_applies_all_operations(mocker):
    synchronizer = make_synchronizer(mocker)
    synchronizer.compare_with_working_set = mocker.Mock(return_value=(["add"], ["update"], ["remove"]))
    add_candidate = mocker.Mock(version="1")
    update_candidate = mocker.Mock(version="2")
    synchronizer.__dict__["candidates"] = {"add": add_candidate, "update": update_candidate}
    update_dist = mocker.Mock(version="1")
    remove_dist = mocker.Mock(version="3")
    synchronizer.working_set = {"update": update_dist, "remove": remove_dist}
    manager = mocker.Mock()
    synchronizer.get_manager = mocker.Mock(return_value=manager)

    synchronizer.synchronize()

    manager.install.assert_called_once_with(add_candidate)
    manager.overwrite.assert_called_once_with(update_dist, update_candidate)
    manager.uninstall.assert_called_once_with(remove_dist)


@pytest.mark.parametrize("already_installed", [False, True])
def test_synchronize_installs_self(mocker, already_installed):
    synchronizer = make_synchronizer(mocker)
    synchronizer.install_self = True
    synchronizer.no_editable = False
    synchronizer.compare_with_working_set = mocker.Mock(return_value=([], [], []))
    synchronizer.__dict__["self_candidate"] = mocker.sentinel.self_candidate
    synchronizer.environment.project.name = "Demo"
    dist = mocker.sentinel.dist
    synchronizer.working_set = {"demo": dist} if already_installed else {}
    manager = mocker.Mock()
    synchronizer.get_manager = mocker.Mock(return_value=manager)

    synchronizer.synchronize()

    if already_installed:
        manager.overwrite.assert_called_once_with(dist, mocker.sentinel.self_candidate)
    else:
        manager.install.assert_called_once_with(mocker.sentinel.self_candidate)



from __future__ import annotations

import pytest

from pdm.exceptions import UninstallError
from pdm.installers.manager import InstallManager


def test_install_passes_candidate_options(mocker):
    environment = mocker.sentinel.environment
    candidate = mocker.Mock()
    candidate.req.editable = False
    candidate.requested = True
    prepared = candidate.prepare.return_value
    prepared.build.return_value = mocker.sentinel.wheel
    prepared.direct_url.return_value = mocker.sentinel.direct_url
    install_wheel = mocker.patch(
        "pdm.installers.manager.install_wheel",
        return_value=mocker.sentinel.dist_info,
    )
    distribution_at = mocker.patch("pdm.installers.manager.Distribution.at")
    manager = InstallManager(environment, use_install_cache=True, rename_pth=True)

    result = manager.install(candidate)

    assert result is distribution_at.return_value
    candidate.prepare.assert_called_once_with(environment)
    install_wheel.assert_called_once_with(
        mocker.sentinel.wheel,
        environment,
        direct_url=mocker.sentinel.direct_url,
        install_links=True,
        rename_pth=True,
        requested=True,
    )
    distribution_at.assert_called_once_with(mocker.sentinel.dist_info)


def test_install_does_not_cache_editable_candidate(mocker):
    candidate = mocker.Mock()
    candidate.req.editable = True
    prepared = candidate.prepare.return_value
    install_wheel = mocker.patch("pdm.installers.manager.install_wheel")
    mocker.patch("pdm.installers.manager.Distribution.at")
    manager = InstallManager(mocker.sentinel.environment, use_install_cache=True)

    manager.install(candidate)

    assert install_wheel.call_args.kwargs["install_links"] is False
    prepared.build.assert_called_once_with()


def test_get_paths_to_remove_uses_environment(mocker):
    environment = mocker.sentinel.environment
    dist = mocker.sentinel.dist
    from_dist = mocker.patch("pdm.installers.manager.StashedRemovePaths.from_dist")
    manager = InstallManager(environment)

    result = manager.get_paths_to_remove(dist)

    assert result is from_dist.return_value
    from_dist.assert_called_once_with(dist, environment=environment)


def test_uninstall_commits_removed_paths(mocker):
    dist = mocker.Mock()
    dist.metadata.get.return_value = "demo"
    manager = InstallManager(mocker.sentinel.environment)
    remove_paths = mocker.patch.object(manager, "get_paths_to_remove").return_value

    manager.uninstall(dist)

    remove_paths.remove.assert_called_once_with()
    remove_paths.commit.assert_called_once_with()
    remove_paths.rollback.assert_not_called()


def test_uninstall_rolls_back_after_remove_error(mocker):
    dist = mocker.Mock()
    manager = InstallManager(mocker.sentinel.environment)
    remove_paths = mocker.patch.object(manager, "get_paths_to_remove").return_value
    error = OSError("cannot remove")
    remove_paths.remove.side_effect = error

    with pytest.raises(UninstallError) as exc_info:
        manager.uninstall(dist)

    assert exc_info.value.__cause__ is error
    remove_paths.commit.assert_not_called()
    remove_paths.rollback.assert_called_once_with()


def test_overwrite_preserves_new_distribution_paths(mocker):
    dist = mocker.Mock()
    dist.metadata.get.return_value = "demo"
    candidate = mocker.sentinel.candidate
    installed = mocker.sentinel.installed
    old_paths = mocker.Mock()
    new_paths = mocker.Mock()
    manager = InstallManager(mocker.sentinel.environment)
    mocker.patch.object(manager, "install", return_value=installed)
    mocker.patch.object(manager, "get_paths_to_remove", side_effect=[old_paths, new_paths])

    manager.overwrite(dist, candidate)

    old_paths.difference_update.assert_called_once_with(new_paths)
    old_paths.remove.assert_called_once_with()
    old_paths.commit.assert_called_once_with()
    old_paths.rollback.assert_not_called()


def test_overwrite_rolls_back_after_remove_error(mocker):
    old_paths = mocker.Mock()
    old_paths.remove.side_effect = OSError("cannot remove")
    manager = InstallManager(mocker.sentinel.environment)
    mocker.patch.object(manager, "install")
    mocker.patch.object(manager, "get_paths_to_remove", side_effect=[old_paths, mocker.Mock()])

    with pytest.raises(UninstallError):
        manager.overwrite(mocker.Mock(), mocker.sentinel.candidate)

    old_paths.commit.assert_not_called()
    old_paths.rollback.assert_called_once_with()
